#==================================================================================================
#Wrapper Module for DFKI Bayes-True Data Generators
#-- This module guarantees explicit parameter logging and replayability of experiments.
#Read demo.py for how this module can be used. --
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================


#==================================================================================================
#INCLUDES
#==================================================================================================
from wgks import *
from wmcg import *
from transforms import *
import shelve
import string
import csv
from scipy.io import *
import scipy
import time
import os
from version import *


#==================================================================================================
#GLOBALS
#==================================================================================================
data_sets = {}
stack = []
transaction_id = 0

#==================================================================================================
#FUNCTIONS
#==================================================================================================

def execute( program=None ):    
    """execute will execute a list of lists (of list of ...) or of dictionaries in order to
    generate data. Resulting data can be found in the global dictionary "data_sets"
    
    Keyword Arguments:
    program:    list of lists or dictionaries.
    """
    global stack
    global data_sets
    if program is None:
        program = stack

    if type( program ) is list:
        for statement in program:
            execute( statement )

    if type( program ) is dict:
        cmd = program.get( "cmd", None )
        if cmd == "wgks":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            k                       = program.get( "no_of_densities", 3 )
            target_error            = program.get( "target_error", 0.1  )
            samples_per_density     = program.get( "samples_per_density", 1000 )
            optimizer               = program.get( "optimizer", "lut" )
            samples_for_optimization= program.get( "samples_for_optimization", 1e+6 )
            
            v,l,p = white_gaussians_on_k_simplex( k, target_error, samples_per_density, optimizer, samples_for_optimization )
            data_sets[name] = { "vectors":v, "labels":l, "probabilities":p}

        if cmd == "k_white":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            k                       = program.get( "no_of_densities", 3 )
            d                       = program.get( "distance", 0.1  )
            samples_per_density     = program.get( "samples_per_density", 1000 )
            v,l,p = k_white_gaussians( d, k, samples_per_density )
            data_sets[name] = { "vectors":v, "labels":l, "probabilities":p }

        if cmd == "solve":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            target_error            = program.get( "target_error", 0.1  )
            total_samples           = program.get( "total_samples", 1000.0 ) #100samples per metric unit
            skeletton               = program["skeletton"]

            r = SkelettonSolution( skeletton, target_error, total_samples )
            data_sets[name] = { "skeletton":r.s, "vectors":r.v, "labels":r.l, "probabilities":r.p, "dims":skeletton.shape[1] }

        if cmd == "embed":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            dims                    = program["dims"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}
            if source_skeletton is not None:
                result["skeletton"] = extend_dimensionality( source_skeletton, dims )
            if source_vectors is not None:
                result["vectors"]   = extend_dimensionality( source_vectors, dims )
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result
            
        if cmd == "rand_scale":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}            
            if source_skeletton is not None:
                result["skeletton"] = source_skeletton * program["random_scaling_vector"]
            if source_vectors is not None:
                result["vectors"]   = source_vectors * program["random_scaling_vector"]
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result
                
        if cmd == "rand_trans":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}            
            if source_skeletton is not None:
                result["skeletton"] = source_skeletton + program["random_translation_vector"]
            if source_vectors is not None:
                result["vectors"]   = source_vectors + program["random_translation_vector"]
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result


        if cmd == "rand_rot":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}
            if source_skeletton is not None:
                result["skeletton"] = scipy.dot( program["random_rotation_matrix"], source_skeletton.T ).T
            if source_vectors is not None:
                result["vectors"]   = scipy.dot( program["random_rotation_matrix"], source_vectors.T ).T
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result

        if cmd == "rand_shear":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}
            if source_skeletton is not None:
                result["skeletton"] = scipy.dot( program["random_shearing_matrix"], source_skeletton.T ).T
            if source_vectors is not None:
                result["vectors"]   = scipy.dot( program["random_shearing_matrix"], source_vectors.T ).T
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result

        if cmd == "recenter":
            source_name             = program["source_name"]  #must raise exceptions if is not there
            target_name             = program["target_name"]  #must raise exceptions if is not there
            new_centroid            = program["new_centroid"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}
            if source_skeletton is not None and source_vectors is None:
                result["skeletton"] = recenter( source_skeletton, new_centroid )
            if source_skeletton is None and source_vectors is not None:
                result["vectors"]   = recenter( source_vectors, new_centroid )
            if source_skeletton is not None and source_vectors is not None:
                off = new_centroid - mean( source_vectors, 0 )[newaxis,:]
                result["vectors"]   = source_vectors + off
                result["skeletton"] = source_skeletton + off
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result

        
        if cmd == "apply_cov":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            result                  =  {}

            if source_skeletton is not None and source_vectors is None:
                result["skeletton"] = apply_cov_matrix( source_skeletton, program["new_sigma"] )[0]
            if source_vectors is None and source_skeletton is not None:
                result["vectors"]   = apply_cov_matrix( source_vectros, program["new_sigma"] )[0]
            if source_vectors  is not None and source_skeletton is not None:
                result["vectors"], result["skeletton"] = apply_cov_matrix( source_vectros, program["new_sigma"], [source_skeletton] ) #source_skeletton will be transformed exactly the same way as source_vectors
            if source_labels is not None:
                result["labels"]    = source_labels
            if source_probabilities is not None:
                result["probabilities"] = source_probabilities
            data_sets[target_name]  = result

        if cmd == "flush":
            export_data_sets( program["what"], program["exporter"] )
            data_sets = {}


#end def execute___________________________________________________________________________________            


def str_matrix( m ):
    """formats a single line python call that can rebuild the matrix via eval()"""
    result = "array( "

    if m.ndim > 1:
        result += "["
        for item in m:
            result += str_matrix( item ) + ","
        result += "]"
    elif m.ndim==1:
        result += "["
        for item in m:
            result += str( item ) + ","
        result += "]"
    return result + ")"
#end def str_matrix________________________________________________________________________________


def age_file( filename ):
    """if filename exists, then it will be renamed to reflect
    that it is not the newest version."""
    name_parts = string.split( filename, "." )
    try:
        index = int( name_parts[-1] )
        name_parts[-1] = str( index+1 )
    except:
        name_parts += ["1"]
    aged_name = string.join( name_parts, "." )

    if os.access( filename, os.R_OK ):
        age_file( aged_name )
        os.system( "mv " + filename + " " + aged_name )
#end def age_file__________________________________________________________________________________
        

def compile2file( source_filename, open_target_filehandle ):
    """copies the python source in source_filename to the already open python file in
    open_target_filehandle. It removes specific lines of code:
    1. from wgks import *
    2. from wmcg import *
    3. from transforms import *
    4. from version import *
    5. from normal_density import *
    6. from optimizers import *
    7.     from lut import *
    """
    try:
        f = open( source_filename, "r" )
    except:
        for path in sys.path:
            try:
                f = open( os.path.join( path, source_filename), "r" )
            except: pass
        try:
            f
        except:
            print "could not find", source_filename
            exit(-1)
            
    for line in f:
        #if line[:-1] in [ "from wgks import *", "from transforms import *", "from wmcg import *", 
        #                "from version import *", "from normal_density import *", "from optimizers import *",
        #                "    from lut import *"]:
        #    continue;        
        s = line.split()
        try:
            if s[0] == "import":
                continue
            if s[2] == "import": 
                continue;
        except:
            pass;
        open_target_filehandle.write( line ); #else write regular stuff
#end def compile2file______________________________________________________________________________


def compile2python( python_file="generation_script.py", program=None, full=1, insert="" ):    
    """compile2python does exactly the same thing as execute() but writes it to a python file.
    Like execute() this function is recursive and recursion will expect python_file to be
    a file handle instead of a string.

    Keyword Arguments:
    program:        None or list of lists or dictionaries. If it is None, then global stack is used.
    python_file:    string or file handle.
    full:           0/1. if full then all necessary generator code is also included
    """

    def is_atomic( program ):
        """checks, whether program (a list object) is containing sublists"""
        for item in program:
            if type( item ) is list:
                return 0
        return 1
    #end def is_atomic_________________________________________________________________________________

    global stack
    global data_sets
    if program is None:
        program = stack

    close_python_file = 0
    if type( python_file ) is str:
        age_file( python_file );
        python_file = open( python_file, "w" )
        python_file.write( "#!/usr/bin/python\n" )
        python_file.write( "#Data description and generators bundled together in this file for publication.\n" )
        python_file.write( "#created on " + time.ctime() + "\n" );
        python_file.write( "#by user: " + str(os.getenv( "USER" )) + "\n" )
        python_file.write( "#in path: " + os.getcwd() + "\n" )
        python_file.write( "#on machine: " + str(os.getenv( "HOSTNAME" )) + "\n" )
        python_file.write( "#generated with " + version + "\n" );
        python_file.write( "#authors " + authors + "\n" );
        python_file.write( "#get from http://paren.iupr.com\n" )

        if full:
            python_file.write( "__name__ = 'deactivate pending self tests'\n\n" )
            python_file.write( "import os, sys\nfrom numpy.random import *\nimport time\nfrom scipy.linalg import *\n" )
            python_file.write( "from time import *\nfrom pylab import *\nfrom numpy import *\nfrom scipy import *\nimport numpy\nimport scipy\n" )
            python_file.write( "from scipy.special import *\nimport csv\n" )
            compile2file( "normal_density.py", python_file )
            compile2file( "lut.py",            python_file )
            compile2file( "optimizers.py",     python_file )
            compile2file( "wgks.py",           python_file )
            compile2file( "wmcg.py",           python_file )
            compile2file( "transforms.py",     python_file )
            compile2file( "version.py",        python_file )
            compile2file( "generators.py",     python_file )

            python_file.write( "\n\n\n#=========================================================================\n" )
            python_file.write( "#THIS IS THE DATA GENERATION SECTION THAT ACTUALLY DOES SOMETHING!\n" )
            python_file.write( "#All of it is generated automatically:\n#=========================================================================\n" )
        else:
            python_file.write( "from generators import *\n" );
        python_file.write( "seed( " + str(int(rand()*1e9)) + " )\n" )
        python_file.write( "class db:\n" )
        python_file.write( "    pass\n\n\n" )

        python_file.write( "#=========================================================================\n" )
        python_file.write( "#THIS IS THE PARALLELIZATION SECTION!\n" )
        python_file.write( "#=========================================================================\n" )
        python_file.write( "VAR_TRANSACTION_ID=-1\n\n")
        python_file.write( "if len( sys.argv ) == 2:\n" )
        python_file.write( "    n = int( sys.argv[1] )\n" )
        python_file.write( "    for i in range(n):\n" ) 
        python_file.write( "        cmd = 'python ' + sys.argv[0]+ ' ' + str( n ) + ' ' + str( i ) + '& '\n" )
        python_file.write( "        print 'calling:', cmd\n" )
        python_file.write( "        os.system( cmd )\n" )
        python_file.write( "    exit(0)\n\n\n" )

        python_file.write( "try:\n" )
        python_file.write( "    CONST_STEP = max(1, int(sys.argv[1]))\n" )
        python_file.write( "    CONST_OFF  = max(0, int( sys.argv[2]))\n" )
        python_file.write( "except:\n" )
        python_file.write( "    CONST_STEP = 1\n" )
        python_file.write( "    CONST_OFF  = 0\n\n" )
        python_file.write( "#=========================================================================\n" )
        python_file.write( "#END OF THE PARALLELIZATION SECTION!\n" )
        python_file.write( "#=========================================================================\n\n\n" )
        close_python_file = 1

    if type( program ) is list:
        if program == []:
            raise Exception( "program cannot be empty!" )

        if is_atomic( program ):
            python_file.write( insert + "VAR_TRANSACTION_ID += 1\n" )
            python_file.write( insert + "if (VAR_TRANSACTION_ID-CONST_OFF)%CONST_STEP==0:\n" )
        else:
            python_file.write( insert + "if 1:\n" ) #this solution is good because indentation preserves list embedding structure and people can manually activate/deactivate portions by setting 1 to 0

        for statement in program:
            compile2python( python_file, statement, insert=insert+"    " )
        
        if close_python_file:
            python_file.write( "\n\n\n#=========================================================================\n" )
            python_file.write( "#END OF ALL FILES\n" )
            python_file.write( "#=========================================================================\n" )
            python_file.close()

    if type( program ) is dict:
        cmd = program.get( "cmd", None )
        if cmd == "wgks":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            k                       = program.get( "no_of_densities", 3 )
            target_error            = program.get( "target_error", 0.1  )
            samples_per_density     = program.get( "samples_per_density", 1000 )
            optimizer               = program.get( "optimizer", "lut" )
            samples_for_optimization= program.get( "samples_for_optimization", 1e+6 )

            data_sets[name] = { "vectors":"dummy", "labels":"dummy", "probabilities":"dummy"}
            python_file.write( insert + name + " = db()\n" )
            python_file.write( insert + name + ".v," + name + ".l," + name + ".p = white_gaussians_on_k_simplex( k=" + str(k) +", target_error=" + str(target_error) + ", samples_per_density=" + str(samples_per_density) + ", optimizer='"+optimizer+"', samples_for_optimization=" + str(samples_for_optimization) +" )\n" )


        if cmd == "k_white":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            k                       = program.get( "no_of_densities", 3 )
            d                       = program.get( "distance", 0.1  )
            samples_per_density     = program.get( "samples_per_density", 1000 )
            python_file.write( insert + name + " = db()\n" )
            python_file.write( insert + name + ".v," + name + ".l," + name + ".p = k_white_gaussians( d=" + str(d) + ",k=" + str(k) + ", samples_per_density=" + str(samples_per_density) + " )\n" )
            data_sets[name] = { "vectors":"dummy", "labels":"dummy", "probabilities":"dummy"}
    

        if cmd == "solve":
            name                    = program.get( "name", "default"+str(len(data_sets)) )
            target_error            = program.get( "target_error", 0.1  )
            total_samples           = program.get( "total_samples", 1000.0 ) #100samples per metric unit
            skeletton               = program["skeletton"] #is a must have here

            python_file.write( insert + name + " = db()\n" )
            python_file.write( insert + "tmp_skeletton = " + str_matrix( skeletton ) + "\n" )
            python_file.write( insert + name + "= SkelettonSolution( skeletton=tmp_skeletton, target_error=" + str(target_error) + ", total_samples=" + str( total_samples ) +" )\n" )
            data_sets[name] = { "skeletton":"dummy", "vectors":"dummy", "labels":"dummy", "probabilities":"dummy"}

        if cmd == "embed":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            dims                    = program["dims"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            if source_skeletton:
                python_file.write( insert + target_name + ".s = extend_dimensionality( " + source_name + ".s, " + str( dims ) + " )\n" ) 
            if source_vectors:
                python_file.write( insert + target_name + ".v = extend_dimensionality( " + source_name + ".v, " + str( dims ) + " )\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]
            
        if cmd == "rand_scale":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            if source_skeletton:
                python_file.write( insert + target_name + ".s = " + source_name + ".s * " + str_matrix( program["random_scaling_vector"] ) + "\n" )
            if source_vectors:
                python_file.write( insert + target_name + ".v = " + source_name + ".v * " + str_matrix( program["random_scaling_vector"] ) + "\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]
                
        if cmd == "rand_trans":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            if source_skeletton:
                python_file.write( insert + target_name + ".s = " + source_name + ".s + " + str_matrix( program["random_translation_vector"] ) + "\n" )
            if source_vectors:
                python_file.write( insert + target_name + ".v = " + source_name + ".v + " + str_matrix( program["random_translation_vector"] ) + "\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]


        if cmd == "rand_rot":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            python_file.write( insert + "rot_matrix = " + str_matrix( program["random_rotation_matrix"] ) + "\n" )
            if source_skeletton:
                python_file.write( insert + target_name + ".s = dot( rot_matrix, " + source_name + ".s.T ).T\n" )
            if source_vectors:
                python_file.write( insert + target_name + ".v = dot( rot_matrix, " + source_name + ".v.T ).T\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]

        if cmd == "rand_shear":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )
            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            python_file.write( insert + "shear_matrix = " + str_matrix( program["random_shearing_matrix"] ) + "\n" )
            if source_skeletton:
                python_file.write( insert + target_name + ".s = dot( shear_matrix, " + source_name + ".s.T ).T\n" )
            if source_vectors:
                python_file.write( insert + target_name + ".v = dot( shear_matrix, " + source_name + ".v.T ).T\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]

        if cmd == "recenter":
            source_name             = program["source_name"]  #must raise exceptions if is not there
            target_name             = program["target_name"]  #must raise exceptions if is not there
            new_centroid            = program["new_centroid"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )

            python_file.write( insert + "if not globals().has_key( '" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )

            if source_skeletton is not None and source_vectors is None:
                python_file.write( insert + target_name + ".s = recenter( " + source_name + ".s, " + str_matrix( new_centroid ) + " )\n" )
            if source_skeletton is None and source_vectors is not None:
                python_file.write( insert + target_name + ".v = recenter( " + source_name + ".v, " + str_matrix( new_centroid ) + " )\n" )
            if source_skeletton is not None and source_vectors is not None:
                python_file.write( insert + "correction_vector = " + str_matrix( new_centroid - mean( source_vectors, 0 )[newaxis,:] ) + "\n" )
                python_file.write( insert + target_name + ".v = " + source_name + ".v + correction_vector\n" )
                python_file.write( insert + target_name + ".s = " + source_name + ".s + correction_vector\n" )
            if source_labels is not None:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities is not None:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]

        if cmd == "apply_cov":
            source_name             = program["source_name"] #must raise exceptions if is not there
            target_name             = program["target_name"] #must raise exceptions if is not there
            source_skeletton        = data_sets[source_name].get( "skeletton", None )
            source_vectors          = data_sets[source_name].get( "vectors", None )
            source_labels           = data_sets[source_name].get( "labels", None )
            source_probabilities    = data_sets[source_name].get( "probabilities", None )

            python_file.write( insert + "if not globals().has_key('" + target_name + "' ):\n    " + insert + target_name + " = db()\n" )
            python_file.write( insert + "new_sigma = " + str_matrix( program["new_sigma"] ) + "\n" )

            if source_skeletton and not source_vectors:
                python_file.write( insert + target_name + ".s = apply_cov_matrix( " + source_name + ".s, new_sigma )[0]\n" )
            if source_vectors and not source_skeletton:
                python_file.write( insert + target_name + ".v = apply_cov_matrix( " + source_name + ".v, new_sigma )[0]\n" )
            if source_vectors and source_skeletton:
                python_file.write( insert + target_name + ".v, " + target_name + ".s = apply_cov_matrix( " + source_name + ".v, new_sigma, [" + source_name + ".s] )\n" )
            if source_labels:
                python_file.write( insert + target_name + ".l = " + source_name + ".l\n" )
            if source_probabilities:
                python_file.write( insert + target_name + ".p = " + source_name + ".p\n" )
            data_sets[target_name] = data_sets[source_name]

        if cmd == "flush":
            if program["exporter"] is numpy_exporter:
                e = "numpy_exporter"
            if program["exporter"] is matlab_exporter:
                e = "matlab_exporter"
            if program["exporter"] is shelve_exporter:
                e = "shelve_exporter"
            if program["exporter"] is csv_exporter:
                e = "csv_exporter"
            if program["exporter"] is nirvana_exporter:
                e = "nirvana_exporter"

            for name in program["what"]:
                python_file.write( insert + "export( description(" + name + "), '" + name + "', " + e + " )\n" )
            if len( program["all"] ) > 0:
                python_file.write( insert + "del " );
                for name in program["all"]:
                    python_file.write( name+", " );
                python_file.write( "\n" )
            data_sets = {}
#end def compile2python____________________________________________________________________________


def save_data_sets( filename = "all_datasets.shelve" ):
    global data_sets
    x = shelve.open( filename, "n" )
    x["data_sets"] = data_sets
    x.close()
#end def save_data_sets____________________________________________________________________________

def restore_data_sets( filename = "all_datasets.shelve" ):
    global data_sets
    x = shelve.open( filename, "c" )
    data_sets = x["data_sets"]
    x.close()
#end def save_data_sets____________________________________________________________________________

def save_stack( filename = "generation_stack.shelve" ):
    global data_sets
    x = shelve.open( filename, "n" )
    x["stack"] = stack
    x.close()
#end def save_data_sets____________________________________________________________________________

def restore_stack( filename = "generation_stack.shelve" ):
    global data_sets
    x = shelve.open( filename, "c" )
    stack = x["stack"]
    x.close()
#end def save_data_sets____________________________________________________________________________


def numpy_exporter( name, s, v, l, p ):
    if s is not None:
        s.dump( name+".skeletton.numpy" )
    if v is not None:
        v.dump( name+".vectors.numpy" )
    if l is not None:
        l.dump( name+".labels.numpy" )
    if p is not None:
        p.dump( name+".probabilities.numpy" )
#end def numpy_exporter____________________________________________________________________________


def shelve_exporter( name, s, v, l, p ):
    x = shelve.open( name+".shelve", "n" )
    if s is not None:
        x["skeletton"] = s
    if v is not None:
        x["vectors"] = v
    if l is not None:
        x["labels"] = l
    if p is not None:
        x["probabilities"] = p
    x.close()
#end def shelve_exporter___________________________________________________________________________


def matlab_exporter( name, s, v, l, p ):
    x = {}
    if s is not None:
        x["skeletton"] = s
    if v is not None:
        x["vectors"] = v
    if l is not None:
        x["labels"] = l
    if p is not None:
        x["probabilities"] = p
    savemat( name+".mat", x )  
#end def matlab_exporter___________________________________________________________________________


def nirvana_exporter( name, s, v, l, p ):
    pass;
#end def nirvana_exporter__________________________________________________________________________


def csv_exporter( name, s, v, l, p ):
    if s is not None:
        csv_writer = csv.writer( open( name+".skeletton.csv", "w" ) )
        csv_writer.writerows( s )
    if v is not None:
        csv_writer = csv.writer( open( name+".vectors.csv", "w" ) )
        csv_writer.writerows( v )
    if l is not None:
        csv_writer = csv.writer( open( name+".labels.csv", "w" ) )
        csv_writer.writerows( l[:,newaxis] )
    if p is not None:
        csv_writer = csv.writer( open( name+".probabilities.csv", "w" ) )
        csv_writer.writerows( p )
#end def csv_eporter_______________________________________________________________________________


def serial_flow_exporter( name, s, v, l, p ):
    rows = len( l );
    try:
        f = open( name + ".csv", "w" );
    except:
        print "serial_flow_exporter(): filename " + name + ".csv could not be opened for write!";
        print "dataset", name, "not saved!";
        return;

    # section
    out = "*" + name + "\n";

    #print header line
    if l is not None:
        out += "L "
    
    if v is not None:
        for col in arange( v.shape[1] ):
            out += "V ";

    if p is not None:
        for col in arange( p.shape[1] ):
            out += "T ";
    out += "\n";

    for row in arange( rows ):
        if l is not None:
            out += str(l[row]) + " ";

        if v is not None:
            for col in arange( v.shape[1] ):
                out += str(v[row,col]) + " ";

        if p is not None:
            for col in arange( p.shape[1] ):
                out += str(p[row,col]) + " ";
        out += "\n";
    f.write( out );
    f.close();
#end def serial_star_exporter

def export_data_sets( list_of_names = None, exporting_function=numpy_exporter ):
    """export_data_sets exports a series of data sets found in data_sets
    into disk files using the exporting_function ( csv_exporter, matlab_exporter, shelve_exporter, numpy_exporter).
    list_of_names can be None: in that case all data sets are taken.
    """
    global data_sets
    if list_of_names is None:
        list_of_names = data_sets.keys()
    
    if type( list_of_names ) is not list:
        list_of_names = [ list_of_names ]

    for item in list_of_names:
        db = data_sets[item]
        s = db.get( "skeletton", None )
        v = db.get( "vectors", None )
        l = db.get( "labels", None )
        p = db.get( "probabilities", None )
        exporting_function( item, s,v,l,p )
#end def export_data_sets__________________________________________________________________________


def export( db, name, exporting_function=numpy_exporter ):
    s = db.get( "skeletton", db.get( "s", None ) )
    v = db.get( "vectors", db.get( "v", None ) )
    l = db.get( "labels", db.get( "l", None ) )
    p = db.get( "probabilities", db.get( "p", None ) )
    exporting_function( name, s,v,l,p )
#end def export_data_sets__________________________________________________________________________


def description( obj ):
    keys = dir( obj )
    result = {}
    for key in keys:
        result[key] = getattr( obj, key )
    return result
#end def description_______________________________________________________________________________


def get_dims( db_name ):
    global data_sets
    try:
        db = data_sets[db_name]
        if db.has_key( "dims" ):
            return db["dims"]
        return max( db.get( "skeletton", scipy.array([[]]) ).shape[1], db.get( "vectors", scipy.array([[]]) ).shape[1] )
    except:
        print "ERROR: data_sets=", data_sets
        print "ERROR: db_name=", db_name
        print "ERROR: db=", db
        return 0    
#end def get_dims__________________________________________________________________________________



#==================================================================================================
#PUBLIC MANIPULATOR INTERFACES
#read demo.py for examples
#==================================================================================================
def CLEAR():
    global stack
    stack = []
#end def___________________________________________________________________________________________

def WGKS( name=None, no_of_densities=3, target_error=0.1, samples_per_density=1000, optimizer="lut", samples_for_optimization=1e5 ):
    global stack
    global data_sets
    stack += [{ "cmd":"wgks", "name":name, "no_of_densities":no_of_densities, "target_error":target_error, "samples_per_density":samples_per_density, "optimizer":optimizer, "samples_for_optimization":samples_for_optimization } ]
    data_sets[name] = {"vectors":"dummy", "labels":"dummy", "probabilities":"dummy", "dims":no_of_densities-1}
#end def___________________________________________________________________________________________


def KWHITES( name=None, no_of_densities=3, distance=1.0, samples_per_density=1000 ):
    global stack
    global data_sets
    stack += [{ "cmd":"k_white", "name":name, "no_of_densities":no_of_densities, "distance":distance, "samples_per_density":samples_per_density } ]
    data_sets[name] = {"vectors":"dummy", "labels":"dummy", "probabilities":"dummy", "dims":no_of_densities-1}
#end def___________________________________________________________________________________________

    
def GCG( name=None, target_error=0.1, no_of_chains=2, chain_lengths=40, no_of_chain_seeds=100, seed_sigma=scipy.eye(2), target_angle=180.0, speed=1.0, total_samples=10000 ):
    global stack
    global data_sets
    skeletton = gaussian_skeletton( no_of_chain_seeds, seed_sigma, no_of_chains, chain_lengths, target_angle, speed )
    stack += [ {"cmd":"solve", "name":name, "target_error":target_error, "total_samples":total_samples, "skeletton":skeletton} ]
    data_sets[name] = {"vectors":"dummy", "labels":"dummy", "probabilities":"dummy", "skeletton":"dummy", "dims":seed_sigma.shape[0]}
#end def___________________________________________________________________________________________


def SCG( name=None, target_error=0.1, radius=2.0, no_of_chains=2, chain_lengths=50, no_of_chain_seeds=100, seed_sigma=scipy.eye(3), target_angle=180.0, speed=1.0, total_samples=10000 ):
    global stack
    global data_sets
    skeletton = gaussian_skeletton( no_of_chain_seeds, seed_sigma, no_of_chains, chain_lengths, target_angle, speed )
    skeletton *= radius/sqrt( sum( skeletton**2, 1 ) )[:,newaxis,:] #normalize
    stack += [ {"cmd":"solve", "name":name, "target_error":target_error, "total_samples":total_samples, "skeletton":skeletton} ]
    data_sets[name] = {"vectors":"dummy", "labels":"dummy", "probabilities":"dummy", "skeletton":"dummy", "dims":seed_sigma.shape[0]}
#end def___________________________________________________________________________________________

            
def WMCG( name=None, target_error=0.1, no_of_chains=2, chain_lengths=15, seed_sigma=scipy.eye(2)*2, target_angle=180.0, speed=1.0, total_samples=10000.0, 
          test_samples=1000, fix_speed=0, fix_angle=0, resample=1 ):
    global stack
    global data_sets
    g = gaussian( no_of_chains, seed_sigma )
    skeletton = grow_wmcg_skeletton( g, chain_lengths, target_angle, speed, test_samples, fix_speed, fix_angle, resample )
    stack += [{ "cmd":"solve", "name":name, "target_error":target_error, "no_of_chains":no_of_chains, 
                "chain_lengths":chain_lengths, "seed_sigma":seed_sigma, "target_angle":target_angle, 
                "speed":speed, "total_samples":total_samples, "test_samples":test_samples, 
                "fix_speed":fix_speed, "fix_angle":fix_angle, "resample":resample, "skeletton":skeletton } ]
    data_sets[name] = {"vectors":"dummy", "labels":"dummy", "probabilities":"dummy", "skeletton":"dummy", "dims":seed_sigma.shape[0]}
#end def___________________________________________________________________________________________


def EMBED( source_name, target_name, dims ):
    global stack
    global data_sets
    stack += [ { "cmd":"embed", "source_name":source_name, "target_name":target_name, "dims":dims } ]
    db = data_sets.get(target_name, data_sets[source_name].copy())
    db["dims"]=dims
    data_sets[target_name] = db
#end def___________________________________________________________________________________________


def RSCALE( source_name, target_name, sigma=1.0 ):
    global stack
    dims = get_dims( source_name )
    scales = random_scaling_vector( dims, sigma )
    stack += [ { "cmd":"rand_scale", "source_name":source_name, "target_name":target_name, "random_scaling_vector":scales } ]
    db = data_sets.get(target_name,  data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________

    
def RMOVE( source_name, target_name, scale=1.0, move_type="gauss" ):
    global stack
    dims = get_dims( source_name )
    if dims <= 0:
        raise Exception( "dims for '" + source_name + "' is " + str(dims) + " and that's a problem. Source does exist?" )
    delta = random_translation_vector( dims, move_type, scale )
    stack += [ { "cmd":"rand_trans", "source_name":source_name, "target_name":target_name, "random_translation_vector":delta } ]
    db = data_sets.get(target_name, data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________

    
def RROT( source_name, target_name ):
    global stack
    dims = get_dims( source_name )
    rot_mat = random_rotation_matrix( dims )
    stack += [ { "cmd":"rand_rot", "source_name":source_name, "target_name":target_name, "random_rotation_matrix":rot_mat } ]
    db = data_sets.get(target_name,  data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________


def RSHEAR( source_name, target_name, sigma_list=[1.0] ):
    global stack
    dims = get_dims( source_name )
    rot_shear = random_shearing_matrix( dims, sigma_list )
    stack += [ { "cmd":"rand_shear", "source_name":source_name, "target_name":target_name, "random_shearing_matrix":rot_shear } ]
    db = data_sets.get(target_name, data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________


def NORMCOV( source_name, target_name, cov_matrix ):
    global stack
    stack += [ { "cmd":"apply_cov", "source_name":source_name, "target_name":target_name, "new_sigma":cov_matrix } ]
    db = data_sets.get(target_name, data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________


def CENTER( source_name, target_name, new_centroid ):
    global stack
    stack += [ { "cmd":"recenter", "source_name":source_name, "target_name":target_name, "new_centroid":atleast_2d( new_centroid ) } ]
    db = data_sets.get(target_name, data_sets[source_name].copy() )
    data_sets[target_name] = db
#end def___________________________________________________________________________________________


def FLUSH( what=None, exporter_function=numpy_exporter ):
    global stack
    global data_sets
    if what is None:
        what = data_sets.keys()
    all_sets = data_sets.keys()
    if type(what) is not list:
        what = [ what ]
    stack += [ { "cmd":"flush", "what":what, "exporter":exporter_function , "all":all_sets} ]
    data_sets = {}
#end def___________________________________________________________________________________________


def TRANSACTION():
    global stack
    transaction = []
    try:
        FLUSH( exporter_function=nirvana_exporter )
        while type(stack[-1]) is dict:
            transaction += [stack.pop(-1)]
    except:
        pass
    finally:
        transaction.reverse()
        if transaction != []:
            stack += [transaction]

#end def___________________________________________________________________________________________
    
    
def random_color():
    symbols = "0248F"
    color = "#"
    for i in arange( 6 ):
        color += symbols[ int32( 4.4 * rand() ) ]
    return color
    


glob_fig = None
def draw1( s, v, l, y_lim=[-10,10], x_lim=[-10,10], wait=3 ):
    """drawt is a DrawTranslator() function replacement"""
    global glob_fig
    ion()
    if glob_fig is None:
        glob_fig = figure(figsize=(15,15))
    glob_fig.clear()
    glob_fig.show()
    d = len(s)
    hold( 'on' )
    color_list = []
    for k in range( d ):
        color_list += [ random_color() ]
        selected = get_indices( int32( l == k ) )
        plot( v[selected,0], v[selected,1], '.', color=color_list[k] )        
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:], color=color_list[k] )
        plot( s[k,0,:], s[k,1,:],'o', color=color_list[k] )
    ylim(y_lim);xlim(x_lim)
    hold( 'off' )
    sleep(wait)


def draw2( v, l, y_lim=[-10,10], x_lim=[-10,10], wait=3 ):
    """drawt is a DrawTranslator() function replacement"""
    global glob_fig
    ion()
    if glob_fig is None:
        glob_fig = figure(figsize=(15,15))
    glob_fig.clear()
    glob_fig.show()
    d = len( unique( l ) )
    hold( 'on' )
    for k in range( d ):
        selected = get_indices( int32( l == k ) )
        plot( v[selected,0], v[selected,1], '.' )        
    ylim(y_lim);xlim(x_lim)
    hold( 'off' )
    sleep(wait)


def VISUALIZE( names=None, y_lim=[-10,10], x_lim=[-10,10], wait=3  ):
    global data_sets

    if names is None:
        names = data_sets.keys()

    if type( names ) == str:
        names = [names]

    for name in names:
        d = data_sets[name]
        if d.has_key( "skeletton" ):
            print "d=", d
            print "s.shape=", d["skeletton"].shape
            draw1( d["skeletton"], d["vectors"], d["labels"], y_lim, x_lim, wait )
        else:
            draw2( d["vectors"], d["labels"], y_lim, x_lim, wait )
        savefig( name+".pdf" )

def EXEC():
    execute()

#==================================================================================================
#END OF FILE
#==================================================================================================

