#==================================================================================================
#Normal Densities Utility Module For Dataset Generators
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================


#==================================================================================================
#INCLUDES
#==================================================================================================
from pylab import *
from scipy import *


def white_gaussian( samples_count, dims ):
    result = zeros( (samples_count, dims) )
    for d in xrange( dims ):
        result[:,d] = normal(0.0, 1.0, size=(samples_count,))
    return result;
#end def white_gaussian____________________________________________________________________________


def normal_distribution( Count, Dimensions, mu, epsilon ):
    mu                  = atleast_2d( mu )
    EigenDescription    = eig( epsilon );                                      #Calculate destortion directions
    Transformation      = EigenDescription[1]  * sqrt(EigenDescription[0]).T;  #make transformation from white gauss to target gauss
    Temp                = white_gaussian( Count, Dimensions );                 #create white normal distribution
    NormalDistribution  = dot( Transformation, Temp.T ) + mu.T;                #apply transformation
    return NormalDistribution.T;                                               #return vector population
#end def normal_distribution_______________________________________________________________________


def normal_likelihood( vectors, center, epsilon_matrix ):
    """Helper function. Computes the likelihood of the density specified through >>inverse_epsilon<< and
    >>sqrt_determinant<< and >>center<< at position >>vector<<. This function
    returns a lieklihood for every vector.
    
    supply >>inverse_epsilon<< and >>sqrt_determinant<< from a central place. This speeds
    up the performance:
    inverse_epsilon  = linalg.inv( Epsilon )
    sqrt_determinant = sqrt( abs( linalg.det( Epsilon ) ) )
    
    Keyword arguments:
    vector -- 2d-numpy-array which can contain a series of vectors (in rows)
    center -- 2d-numpy-array. vector containing the center position of the gaussian
    inverse_epsilon -- 2d--numpy-array. matrix with the inverse of the cov matrix
    sqrt_determinant -- float. square root of absolute determinant of the cov matrix
    """

    center = atleast_2d( center );
    sqrt_determinant = sqrt( linalg.det( epsilon_matrix ) );
    inverse_epsilon = linalg.inv( epsilon_matrix );

    j = vectors - center;    
    d = len( center );
    b = transpose(exp( -0.5 * sum(j.T * dot(inverse_epsilon, j.T), 0) ) / ((2 * pi)**(d / 2.0) * sqrt_determinant));
    return b;
#end def normal_likelihood_________________________________________________________________________


def normal_likelihood_optimized( vectors, center, inv_sigma, sqrt_det_sigma ):
    """Helper function. Computes the likelihood of the density specified through >>inverse_epsilon<< and
    >>sqrt_determinant<< and >>center<< at position >>vector<<. This function
    returns a lieklihood for every vector.
    
    supply >>inverse_epsilon<< and >>sqrt_determinant<< from a central place. This speeds
    up the performance:
    inverse_epsilon  = linalg.inv( Epsilon )
    sqrt_determinant = sqrt( abs( linalg.det( Epsilon ) ) )
    
    Keyword arguments:
    vector -- 2d-numpy-array which can contain a series of vectors (in rows)
    center -- 2d-numpy-array. vector containing the center position of the gaussian
    inverse_epsilon -- 2d--numpy-array. matrix with the inverse of the cov matrix
    sqrt_determinant -- float. square root of absolute determinant of the cov matrix
    """

    center = atleast_2d( center );
    sqrt_determinant = sqrt_det_sigma;
    inverse_epsilon = inv_sigma;

    j = vectors - center;  
    d = len( center );
    b = transpose(exp( -0.5 * sum(j.T * dot(inverse_epsilon, j.T), 0) ) / ((2 * pi)**(d / 2.0) * sqrt_determinant));
    return b;
#end def normal_likelihood_optimized_______________________________________________________________


#==================================================================================================
#END OF FILE
#==================================================================================================

