#==================================================================================================
#Optimizer Methods For Dataset Generators
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================


#==================================================================================================
#INCLUDES
#==================================================================================================
from numpy import *
from scipy import *
import sys


#==================================================================================================
#FUNCTIONS
#==================================================================================================

def descent( eval_function, initial_values, tau=0.15,
            max_runs=1000, epsilon=0.001, scale_up = 1.5, base = 0.7,
            initial_rate = 1.0, healing_rate = 0.9, step_limit=1.0, heals=10 ):
    """descent implements a gradient descent with a few heuristics for learning rate adaptation.
    This function returns a vector as optimal result.

    Keyword Arguments:
    eval_function:          function handle callable by the form   f(numpy.array(v))
    initial_values:         1d array. Starting position in optimization space.
    tau:                    float>0. This value specifies the numerical delta for derivative computation
    max_runs:               int>>1. Maximum number of optimization epochs.
    epsilon:                float>0. Constant for early termination.
    scale_up:               float>1.0. Maximum scale up for the learning rate for each epoch
    base:                   float[-1,1]. Cosine of weight change angle where learning rate is not adapted
    initial_rate:           float>0. Initial learning rate
    healing_rate:           float[0,1]. If obscure numerical problems occur, then make the optimization parameters less aggressive by this factor
    step_limit:             float>0. Avoid fly-away effects. Maximum step vector norm.
    heals:                  int>=0. Number of healing attempts before it gives up.
    """
    good = 0;
    position = copy( initial_values );
    dims = len( position );
    deltas = eye( dims );
    adaptation = 1.0;
    print "INITIALIZING NUMERICAL DESCENT --- SCALE_UP:", scale_up, "BASE:",base, "POS:", position
    while not good and heals>0:
        lr = initial_rate
        delta_old = zeros( position.shape );
        old_score = eval_function( position );
        good = 1; #assume it's good
        for i in xrange( max_runs ):
            #optimizer section
            old_position = copy( position );
            #print "DEBUG position", position
            #print "DEBUG deltas", deltas
            gradient = zeros( dims );
            for dim in xrange( dims ):
                #print "DEBUG position+", position + deltas[dim]*tau
                #print "DEBUG position-", position - deltas[dim]*tau
                #print position, deltas[dim]*tau
                gradient[dim] = (eval_function( position + deltas[dim]*tau ) - eval_function( position - deltas[dim]*tau ))/(2*tau)
            #print "\rDEBUG LR:",lr, "GRAD:", gradient, "+d:", eval_function( position + deltas[dim]*tau ), "-d:", eval_function( position - deltas[dim]*tau ),
            position -= lr * gradient;
            delta_new = position-old_position;
            score = eval_function( position );
            try:
                #section for LR adaptation
                DIR = float(dot( delta_new/(linalg.norm(delta_new)+1e-308), delta_old/(linalg.norm(delta_old)+1e-308) ))
                #what does the next line do: if step length changes somehow unexpected from what the lr suggest, then become cautious!
                #print "old:",linalg.norm(delta_old),"new:",linalg.norm(delta_new), "lr:",lr
                DIR *= min((linalg.norm(delta_old)*adaptation)/linalg.norm(delta_new), linalg.norm(delta_new)/(adaptation*linalg.norm(delta_old)) ); #the concept of direction doesn't work well in 1D


                delta_old = delta_new;
                if i > 0:
                    adaptation = float(exp( (DIR-base)/((1.0-base)*(1.0/(log(scale_up))))))
                else: adaptation=1.0


                print "\rEPOCH", i, "ERROR:", score, "LR:", lr, "G:", gradient, "POS:", position, "DIR:", DIR, "ADAPT:", adaptation,
                sys.stdout.flush()

                if linalg.norm( gradient ) < epsilon or lr < epsilon:
                    print "EARLY STOP DUE TO EPSILON"
                    return position;
            except Exception, e:
                print e
                #new numpy style healing
                good = 0;
                scale_up = (scale_up-1.0)*healing_rate+1.0;
                base = 1.0 - (1.0-base)*healing_rate;
                initial_rate *= healing_rate;
                heals -= 1;
                #tau *= healing_rate;
                position = copy( initial_values );
                print "\nSCALE_UP:", scale_up, "BASE:",base, "initial_rate:", initial_rate, "POS:", position
                print "HEALING"
                break;

            lr *= adaptation;
            old_score = score;
    print "LATE STOP DUE TO MAX_RUNS"
    return position;
#end def descent___________________________________________________________________________________


def descent_with_analytical_gradient( eval_function, eval_function_derivation, initial_values, 
                                        max_runs=1000, epsilon=0.0001, scale_up = 1.5, base = 0.7,
                                        initial_rate = 1.0, healing_rate = 0.9, step_limit=1.0, heals=10 ):
    """descent implements a gradient descent with a few heuristics for learning rate adaptation.
    This function returns a vector as optimal result.

    Keyword Arguments:
    eval_function:          function handle callable by the form   f(numpy.array(v))
    eval_function_derivation: function handle callable by the form   f(numpy.array(v))
    initial_values:         1d array. Starting position in optimization space.
    tau:                    float>0. This value specifies the numerical delta for derivative computation
    max_runs:               int>>1. Maximum number of optimization epochs.
    epsilon:                float>0. Constant for early termination.
    scale_up:               float>1.0. Maximum scale up for the learning rate for each epoch
    base:                   float[-1,1]. Cosine of weight change angle where learning rate is not adapted
    initial_rate:           float>0. Initial learning rate
    healing_rate:           float[0,1]. If obscure numerical problems occur, then make the optimization parameters less aggressive by this factor
    step_limit:             float>0. Avoid fly-away effects. Maximum step vector norm.
    heals:                  int>=0. Number of healing attempts before it gives up.
    """
    good = 0;
    position = copy( initial_values );
    dims = len( position );
    deltas = eye( dims );
    adaptation=1.0;
    print "INITIALIZING --- SCALE_UP:", scale_up, "BASE:",base, "POS:", position
    while not good and heals>0:
        lr = initial_rate
        delta_old = zeros( position.shape );
        old_score = eval_function( position );
        good = 1;
        for i in xrange( max_runs ):
            #optimizer section
            old_position = copy( position );
            gradient = eval_function_derivation( position );
            position -= minimum( step_limit, maximum(-step_limit, lr * gradient )); #empiric hack
            delta_new = position-old_position;
            score = eval_function( position );

            #section for LR adaptation
            try:
                DIR = float(dot( delta_new/(linalg.norm(delta_new)+1e-308), delta_old/(linalg.norm(delta_old)+1e-308) ))
                #what does the next line do: if step length changes somehow unexpected from what the lr suggest, then become cautious!
                DIR *= min((linalg.norm(delta_old)*adaptation)/linalg.norm(delta_new), linalg.norm(delta_new)/(adaptation*linalg.norm(delta_old)) ); #the concept of direction doesn't work well in 1D
                delta_old = delta_new;
                if i > 0:
                    adaptation = exp( (DIR-base)/((1.0-base)*(1.0/(log(scale_up)))))
                else: adaptation=1.0


                if linalg.norm( gradient ) < epsilon or lr < epsilon:
                    print "EARLY STOP DUE TO EPSILON"
                    return position;
                else: 
                    print "\rEPOCH", i, "ERROR:", score, "LR:", lr, "G:", gradient, "POS:", position, "DIR:", DIR, "ADAPT:", adaptation,
                    sys.stdout.flush()
            except:
                #new numpy style healing
                good = 0;
                scale_up = (scale_up-1.0)*healing_rate+1.0;
                base = 1.0 - (1.0-base)*healing_rate;
                initial_rate *= healing_rate;
                heals -=1;
                position = copy( initial_values );
                print "SCALE_UP:", scale_up, "BASE:",base
                print "HEALING"
                break;


            #old style numpy self healing
            if isnan(score) or isnan(lr) or isnan(gradient) or isnan(float(position)) or isnan(DIR) or isnan(adaptation):
                good = 0;
                scale_up = (scale_up-1.0)*healing_rate+1.0;
                base = 1.0 - (1.0-base)*healing_rate;
                heals -=1;
                position = copy( initial_values );
                print "SCALE_UP:", scale_up, "BASE:",base
                print "HEALING"
                break;

            lr *= adaptation;
            old_score = score;
    print "LATE STOP DUE TO MAX_RUNS"
    return position;
#end def descent_with_analytical_gradient__________________________________________________________


def setup_simplex( n, distance ):	
    """Helper function. Creates a numpy array with coordinates of a n-simplex where any edge is >>distance<< long. 
    The return value is a 2d-numpy-array.

    Keyword arguments:
    n -- int. number of densities
    distance -- float. distance apart for densities
    """
    
    if n <= 1:
        return array([[0]])
    
    result = zeros((n, n-1))
    sub_simplex = setup_simplex(n-1, distance)
    base_foot = sub_simplex.sum(0) / (n - 1)
    r = linalg.norm(base_foot)
    y = sqrt(distance**2 - r**2)
    result[0:n-1,0:n-2] = sub_simplex
    result[n-1,0:n-2] = base_foot
    result[n-1,n-2] = y
    return result
#end def setup_simplex_____________________________________________________________________________


def amoeba( eval_function, dims, initial_values=None, gamma=2.0, beta=0.5, sigma=0.5, max_runs=1000, e1=0.000001, e2=0.000001, e3=0.000001, abort_under_cost=None ):
    """amoeba is a downhill simplex optimizer as described in Nelder and Mead, 1965.
    amoeba is well suited for continuous problems but it can get stuck in a local minimum, just like any other local method.

    Keyword arguments:
    eval_function -- function. fitnes function must take a vector and return a scalar value
    dims -- int. what is the number of dimensions to optimize in
    initial_values -- 2d numpy array or None. starting shape of the simplex. it must fit the size (dims+1)x(dims)
    gamma -- float. expansion coefficient (should of course be always > 1.0 )
    betta -- float. inner kontraction coefficient. (between 0.0 and 1.0)
    sigma -- float. total kontraction coefficient. (between 0.0 and 1.0)
    max_runs -- int. maximal number of optimization steps. (>0)
    e1 -- float. if average of function values changes less then e1, then stop
    e2 -- float. if the newly set vertex is closer to simplex than e2, then stop
    e3 -- float. if the positioning of the new vertex improved the best fitness score less then e3, then stop
    """

    def update_all_values( simplex, dims, eval_function ):
        """ helper function defined internally because this part is used twice """
        for i in range( dims+1 ):
            simplex[i,0] = eval_function( simplex[i,1:] );
        #END OF FUNCTION

    #BEGIN AMOEBA
    simplex = zeros( ( dims+1, dims+1 ) );
    if initial_values is not None:
        simplex[:,1:] = initial_values;
    else:
        simplex[:,1:] = setup_simplex( dims+1, 1.0 );

    update_all_values( simplex, dims, eval_function );

    correction2 = 1.0 / ( dims + 1 );
    correction1 = (dims+1.0) / dims
    for epoch in range( max_runs ):
        #print simplex
        ref = argmax( simplex[:,0] );    
        refvec = simplex[ref,1:].copy();
        center = mean( concatenate( ( simplex[0:ref,1:], simplex[ref+1:,1:]), 0 ), 0 )
        #print "center:", center
        #plot( simplex[:,1], simplex[:,2], 'x' );
        #plot( array([center[0]]), array([center[1]]), 'x' ); xlim( [ -3, 10]); ylim( [-3,10]); show();

        deltavec = center-refvec;
        regular_flip_vec = center + deltavec;
        regular_flip_fit = eval_function( regular_flip_vec );
        
        SortedValues = simplex[:,0].copy();
        SortedValues.sort();
        #print SortedValues
        MinFit = SortedValues[0];
        MaxFit = SortedValues[-1];
        LesFit = SortedValues[-2];
        MinFit = min( simplex[:,0] );
        MaxFit = max( simplex[:,0] );
        CurFit = mean( simplex[:,0] );

        #print "Min", MinFit, "Max", MaxFit, "Les", LesFit
        print "\rEPOCH",epoch,":",
        if regular_flip_fit < MinFit:
            #EXPANSION
            expanded_flip_vec = center + gamma * deltavec;
            expanded_flip_fit = eval_function( expanded_flip_vec );

            if expanded_flip_fit < regular_flip_fit:
                #ONLY FLIP
                print "ONLY FLIP 1",
                simplex[ref,1:] = expanded_flip_vec;
                simplex[ref,0] = expanded_flip_fit;
            else:
                #FLIP AND EXPAND
                print "FLIP AND EXPAND",
                simplex[ref,1:] = regular_flip_vec;
                simplex[ref,0] = regular_flip_fit;
        elif regular_flip_fit >= LesFit and regular_flip_fit < MaxFit:
            #OUTER KONTRACTION
            contracted_flip_vec = center + beta * deltavec;
            contracted_flip_fit = eval_function( contracted_flip_vec );            

            if contracted_flip_fit < regular_flip_fit:
                #FLIP AND CONTRACT
                print "FLIP AND CONTRACT",
                simplex[ref,1:] = contracted_flip_vec;
                simplex[ref,0] = contracted_flip_fit;
            else:
                #SCHRINK ALL SIDES
                print "SHRINK ALL SIDES 1",
                simplex[:,1:] = ( simplex[:,1:] - center ) * sigma + center;
                update_all_values( simplex, dims, eval_function );
        elif regular_flip_fit >= MaxFit:
            #INNER CONTRACTION
            contracted_vec = center - beta * deltavec;
            contracted_fit = eval_function( contracted_vec );            
            
            if contracted_fit < MaxFit: # and contracted_fit < regular_flip_fit:
                #CONTRACT BUT DO NOT FLIP
                print "CONTRACT BUT DO NOT FLIP",
                simplex[ref,1:] = contracted_vec;
                simplex[ref,0] = contracted_fit;
            else:
                #SCHRINK ALL SIDES
                print "SHRINK ALL SIDES 2",
                simplex[:,1:] = ( simplex[:,1:] - center ) * sigma + center;
                update_all_values( simplex, dims, eval_function );
        else:
            #JUST FLIP A BIT
            print "ONLY FLIP 2",
            simplex[ref,1:] = regular_flip_vec;
            simplex[ref,0] = regular_flip_fit;

        if mean( SortedValues ) < e1:
            print "STOP DUE TO AVERAGE FUNCTION CHANGE"
            return center;

        if linalg.norm( refvec - simplex[ref,1:] ) < e2:
            print "STOP DUE TO TOO SMALL SIMPLEX CHANGE"
            return center;

        if abs( MaxFit - simplex[ref,0] ) < e3:
            print "STOP DUE TO MARGINAL IMPROVEMENT"
            return center;

        if abort_under_cost is not None:
            if CurFit < abort_under_cost:
                print "STOP DUE TO QUITE OPTIMAL COST"
                return center

        sys.stdout.flush()
    return center;
#end def amoeba____________________________________________________________________________________


#==================================================================================================
#END OF FILE
#==================================================================================================

