#==================================================================================================
#Random Transformations Module
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================


#==================================================================================================
#INCLUDES
#==================================================================================================
from wgks import *


#==================================================================================================
#FUNCTIONS
#==================================================================================================


def random_rotation_matrix( dimensions ):
    """ Creates a random rotation matrix in >>dimensions<< dimensions.
    RETURNS Array of a random rotation matrix

    Keyword Arguments:    
    dimensions:	No. of (external) dimensions for the rotation matrix
    """
    A   = white_gaussian( dimensions, dimensions );
    Q,R = linalg.qr(A);
    sgn = sign(rand()-.5);    
    return sgn*Q;
#end def random_rotation_matrix____________________________________________________________________


def random_translation_vector( dimensions, type = "uniform", scale = 1.0 ):
    """Creates a random vector for translations in >>dimensions<< dimensions.
    RETURNS Array of a random translation vector
    
    Keyword Arguments:
    dimensions:	No. of (external) dimensions for the translation vector
    type:          Either "uniform" or "gauss"
    scale:         scale factor for the maximal Translation respectively standard deviation when chosen 
                   Gaussian
    """
    if type == "uniform":
        Translation = rand( dimensions ) * scale;
    elif type == "gauss":
        Translation = normal(0., scale, dimensions);
    else:
        print "ERROR: type has to be either 'unif' or 'gauss'. Zero translation vector has been returned!";
        Translation = zeros(dimensions);
    #end if
    
    return Translation[newaxis,:];
#end def random_translation_vector_________________________________________________________________


def random_scaling_vector(dimensions, sigma=1.0 ):
    """Creates a random scaling vector in >>dimensions<< dimensions with sigma >>sigma<<.
    RETURNS Array of a random scaling matrix

    Keyword Arguments:
    dimensions:	No. of (external) dimensions for the scaling matrix
    sigma:         standard deviation as measure for the factor of scaling
    """
    return exp( normal(0., sigma, dimensions) )[newaxis,:]
#end def rand_scaling_vector_______________________________________________________________________


def random_shearing_matrix( dimensions, sigma_list = [1.]):
    """Creates a random shearing matrix for len(sigma_list) dimensions in a >>dimensions<< dimensional space.

    Keyword Arguments:
    dimensions:	No. of (external) dimensions for the scaling matrix
    sigma_list:    list standard deviations as measure for the shearing factor; the length of the list
                   determines the number of shearing operations (has to be less than n*(n-1) for n dimensionality
    """

    #build index list
    possible_indices = [];
    for i in range(dimensions):
        for j in range(dimensions):
            if j != i:
                possible_indices += [(i,j)];
            #end if
        #end for
    #end for
    shearing_indices = permutation(possible_indices)[:len(sigma_list)];
    
    #build shearing matrix
    shearing_matrix = diag(ones(dimensions));
    for k in range(len(sigma_list)):
        i = shearing_indices[k][0];
        j = shearing_indices[k][1];
        
        shearing = normal(0., sigma_list[k]);
        shearing_matrix[i][j] = shearing;
    #end for
    
    return shearing_matrix;
#end def random_shearing_matrix____________________________________________________________________


def extend_dimensionality( vectors, target_dimensionality ):
    """extend_dimensionality embeds vectors into a higher dimensional space.
    
    Keyword Arguments:
    vectors:                    2d array. vectors are in rows
    target_dimensionality:      int>=dim(vecs). number of dimensions of the high dimensional space
    """
    if vectors is None:
        return None
    result = zeros( ( vectors.shape[0], target_dimensionality ) )
    result[:,:vectors.shape[1]] = vectors
    return result
#end def extend_dimensionality_____________________________________________________________________


def transform_by_matrix( vex, new_coordinate_system ):
    """applies regular affine transforms"""
    return dot( new_coordinate_system, vex.T ).T
#end def transform_by_matrix_______________________________________________________________________


def transform_by_cov( vecs, epsilon ):
    """puts a covariance on top"""
    EigenDescription    = eig( epsilon );                                      #Calculate destortion directions
    Transformation      = EigenDescription[1]  * sqrt(EigenDescription[0]);    #make transformation from white gauss to target gauss
    NormalDistribution  = dot( Transformation, vecs.T ).T;                     #apply transformation
    return NormalDistribution;                                                 #return vector population
#end def transform_by_cov__________________________________________________________________________


def apply_cov_matrix( vecs, new_sigma, related_vector_sets = [] ):
    """changes the sigma of the data, but not its centroid"""
    center = mean( vecs, 0 )[newaxis,:]
    vecs -= center
    vs, vv = eig(cov( vecs.T ))
    M1 = linalg.inv((vv*sqrt(vs)))
    whitened_vecs = dot( M1, vecs.T ).T
    vs, vv = eig( new_sigma )
    M2 = (vv*sqrt(vs))
    final_vecs = dot( M2, whitened_vecs.T ).T + center

    transformed = [ final_vecs ]
    for vecs in related_vector_sets:
        vecs -= center
        whitened_vecs = dot( M1, vecs.T ).T
        final_vecs = dot( M2, whitened_vecs.T ).T + center
        transformed += [final_vecs]
    return transformed


#end def apply_cov_matrix__________________________________________________________________________


def recenter( vecs, new_center ):
    center = atleast_2d( new_center )
    return (vecs-mean(vecs,0))+center
#end def recenter


#==================================================================================================
#SELF TEST
#==================================================================================================
if __name__ == "__main__":
    print "3D random rotation matrix: ";
    print RandRotation(3);
    print "3D random uniform translation: ";
    print RandTranslation(3, "unif", 5.);
    print "3D random scaling matrix: ";
    print RandScaling(3, 5.);
    print "3D random shearing matrix: ";
    print RandShearing(3, [.5, 1]);
    
    #debug
    #print S;
    #Samples = rand(1000,3);
    #T1 = Translator(Samples);
    #T1.SetLabel("old")
    #T2 = Translator(dot(S, Samples.T).T);
    #T2.SetLabel("new")
    #figure()
    #DrawTranslator(T1 + T2,[0,1]);
    #figure()
    #DrawTranslator(T1 + T2,[0,2]);
    #figure()
    #DrawTranslator(T1 + T2,[1,2]);
    #show();
#end if

#==================================================================================================
#END OF FILE
#==================================================================================================

