#==================================================================================================
#White Gaussians on Chain Generator WGKS
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================
#
#==================================================================================================
#OVERVIEW OVER FUNCTIONAL DEPENDENCIES
#==================================================================================================
#
#                                white_gaussians_on_k_simplex()
#                                              |
#      +-------------------------+-------------+-----------+----------------------------------+
#      |                         |                         |                                  |
#edge_length_of_k_simplex_lut()  |       edge_length_of_k_simplex_numerical_descent()     k_white_gaussians()
#      |                         |                         |                                  |
#      +----------------+        |                         |                                  |
#      |                |        |                         |                                  |
# extrapolate_y_shift() | edge_length_of_k_simplex_amoeba()|                                 /|
#                       |        |            +------------+                                / |
#                       |        +-----------/------+      |                               /  |
#                       |        |          /        \     |                              /   |
#                       +---adapt_error_function()   FastWGKSErrorModel                  /   position_whites()
#                       |        |                         |                            /     |
#                       +--modulated_erro_function()       +----------setup_simplex()--+      +--------------+
#                                |                                                            |              |
#                              erfcinv()                                                white_gaussian()     |
#                                |                                                                  gauss_likelihood()
#                              ltqnorm()
#
#   TRY_LUT
#      |
#      +--------------------------------+
#      |                                |
#   reproduce_lookup_table()       secure_lut()
#
#
#    TEST_CASES
#      |
#    sort_rows()
#
#==================================================================================================


#==================================================================================================
#INCLUDES
#==================================================================================================
from math import *
from pylab import *
from scipy import *
from scipy.special import *
from time import *
from optimizers import *
from normal_density import *


class FastWGKSErrorModel:
    """This class is a cookie class for the optimization, as the optimizers do
    not provide means of passing constants to the cost function.
    """

    def __init__( self, k, target_error, samples=50000 ):
        """This is the constructor that stores and pregenerates constants for the cost_function

        Keyword Arguments:
        k:              int>1. Number of densities
        target_error:   float[0,1]. Wished Bayes error rate.
        samples:        int>>1. Number of samples per density
        """
        vertices = setup_simplex( k, 1.0 );
        center = vertices.mean( 0 );  
	print "k=",k, "error=", target_error
	print "vertices:\n",vertices
	print "center m0:\n", center, "length of m0:", norm( center )
        self.my_recovery_ratio = 1.0 / linalg.norm( center );
        self.my_dims = k - 1;
        self.my_S = (center - vertices[1:,:])/sqrt(((center - vertices[1:,:])**2).sum(1)) #this will create a coordinate system matrix
	print vertices[0,:].shape, center.shape
	self.my_head_vec = vertices[0,:] - center
        self.my_S_inverse = linalg.inv( self.my_S )
	print "my_S\n", self.my_S
	print "inv my_S\n", self.my_S_inverse
	print "head_vec\n", self.my_head_vec
	print "S*head_vec\n", dot( self.my_S, self.my_head_vec[:,newaxis] )
	exit(0);
        self.my_sigma = dot( self.my_S_inverse.T, self.my_S_inverse )
        self.my_samples = normal_distribution( samples, self.my_dims, zeros(self.my_dims), self.my_sigma )
        self.my_target_error = target_error;
    #END DEF __INIT__
        

    def bayes_error( self, l_dash ):
        """Computes the Bayes error by 1st. octant discounting"""
        samples = copy( self.my_samples ) + ones( self.my_dims ) * l_dash
	print "OPTIMIZER DEBUG OUTPUT:"
	print "length of vec l=", norm( ones( self.my_dims ) * l_dash ), l_dash
	print "length of Sinv * vel l=", norm( dot(self.my_S, (-ones( self.my_dims ) * l_dash)[:,newaxis]) )
	print "should be:", self.my_recovery_ratio * l_dash
        accuracy =  mean(( sign( samples ) > 0 ).all( 1 ));
        return 1.0 - accuracy;
    #END DEF bayes_error


    def convert_2_true_edge_length( self, l_dash ):
        return self.my_recovery_ratio * l_dash
    #END DEF convert_2_true_edge_length


    def cost_function( self, v ):
        return ( self.my_target_error - self.bayes_error( v ) )**2
    #END DEF cost_function    
#END DEF CLASS FastWGKSErrorModel__________________________________________________________________



def reproduce_lookup_table( optimizer="amoeba", samples_for_optimization=1e6, average_over=50, grow=1.15, repeat=34 ):
    """reproduce_lookup_table is the function that reproduces the look up table for the WGKS. Normally this 
    is not used.

    Keyword Arguments:
    optimizer:                  string ("amoeba","descent").
    samples_for_optimization:   int>>>1.
    average_over:               int>>1.
    grow:                       float>1.0. Growth base for test distribution
    repeat:                     int>>1. How many growth operations shall be performed.
    """
    dims = unique(int32(grow**arange(repeat)))[1:]; #never take 1 and do not repeat anything twice
    total = average_over*len(dims)
    factor = 1.0 / average_over;
    counter = 0
    t1 = time();
    table = {}
    #repeat estimations by optimizer for average_over times:
    for run in arange( average_over ):
        for dim in dims: #estimate for every support dimension: dimensions in between are not estimated.
            print "K=", dim, "RUN=",run
            if dim == 2:
                table[dim]=0.0;
            else:
                error_model = FastWGKSErrorModel( dim, 0.5, samples_for_optimization ) #samples_per_density*k );  #optimize for 50% error
                y = amoeba( error_model.cost_function, 1 )[0]
                table[dim] = table.get( dim, 0.0 ) + factor * error_model.convert_2_true_edge_length( y )
            P = counter*100.0/total
            if P==0:
                ETA = "???"
            else:
                ETA = 100*((time()-t1)/P)-(time()-t1)
            print "PROGRESS:%2.2f"%P+"% ETA:", ETA, "s"
            counter+=1
        
    #now it is neccessary to compact and interpolate the values into a look up table:
    max_dim = max(dims+1)
    result_dims   = arange( 2, max_dim )
    result_values = zeros( max_dim-2 )
    keys = table.keys()
    last_dim = keys.pop(0)
    for d in result_dims:
        if d==last_dim:
            result_values[d-2] == table[d]
        if d>last_dim and d<keys[0]:
                result_values[d-2] = ((d-last_dim)*(table[keys[0]]-table[last_dim])/(keys[0]-last_dim))+table[last_dim] #interpolate
        if len(keys)>0 and d==keys[0]:
                last_dim = keys.pop(0)
                result_values[d-2] = table[d]
        if len(keys)==0:
                result_values[d-2] = table[last_dim]
    result = {}
    print "result_dims:", result_dims
    print "result_values:", result_values
    for d in result_dims:
        result[d] = result_values[d-2]    
    return result;
#END DEF reproduce_lookup_table____________________________________________________________________


def secure_lut( table, filename="lut.py" ):
    """saves the lut to a file.

    Keywords Arguments:
    table:          dictionary. Keys are ints 2..100 and values are s_k floats.
    fileanme:       string. Path to lut.py
    """
    f = open( filename, "w" );
    f.write( "Table_y_shift = " +str(table).replace(",",",\n")+"\n" );
    f.close();
#END DEF secure_lut________________________________________________________________________________




#In this TRY we try to load lut.py. If it's not there, then lut needs to get regenerated.
try:
    #UNCOMMENT THE NEXT LINE IF YOU WANT TO USE FALLBACK SOLUTION
    #raise Exception
    from lut import *
    for i in arange(2,100):
        if not Table_y_shift.has_key( i ):
            print "PROBLEM WITH lut.py: MISSING KEYS"
            raise Exception
        if i>2 and Table_y_shift[i] < Table_y_shift[i-1]:
            print "PROBLEM WITH lut.py: BAD VALUES!"
            print "i:", i-1, "val:", Table_y_shift[i-1]
            print "i:", i, "val:", Table_y_shift[i]
            raise Exception 
except: #if lut is not there then:
    print "Do you want to use default table [D] or do you want to regenerate the table[R]?",
    answer = sys.stdin.readline();
    if answer[0].upper() == "D":
        #FALLBACK SOLUTION FROM PAPER
        Table_y_shift = {2: 0., \
        3: 0.78545073168099999, \
        4: 1.18614011249, \
        5: 1.43909624827, \
        6: 1.6308353824199999, \
        7: 1.77032540369, \
        8: 1.8869956481300001, \
        9: 2.0057598688099998, \
        10: 2.0834668720799998, \
        11: 2.15502439144, \
        12: 2.2265819107899998, \
        13: 2.2879206293699998, \
        14: 2.3492593479599999, \
        15: 2.3981006331599999, \
        16: 2.4469419183599999, \
        17: 2.4835660326100002, \
        18: 2.5201901468600001, \
        19: 2.5589019888700002, \
        20: 2.5976138308699999, \
        21: 2.62650311209, \
        22: 2.6553923933000001, \
        23: 2.6842816745200002, \
        24: 2.7131709557299999, \
        25: 2.74206023695, \
        26: 2.7647745635800001, \
        27: 2.7874888902000001, \
        28: 2.8102032168300002, \
        29: 2.8329175434499998, \
        30: 2.8556318700799999, \
        31: 2.8745581194300001, \
        32: 2.8934843687799998, \
        33: 2.91241061813, \
        34: 2.9313368674800002, \
        35: 2.95026311683, \
        36: 2.9653162365100001, \
        37: 2.9803693562000002, \
        38: 2.9954224758799999, \
        39: 3.01047559557, \
        40: 3.0255287152500001, \
        41: 3.04045652276, \
        42: 3.0553843302699999, \
        43: 3.0703121377899998, \
        44: 3.0852399453000001, \
        45: 3.10016775281, \
        46: 3.1115333336900002, \
        47: 3.1228989145699999, \
        48: 3.1342644954400001, \
        49: 3.1456300763199998, \
        50: 3.1569956572, \
        51: 3.1644660956199999, \
        52: 3.1719365340399999, \
        53: 3.1794069724599998, \
        54: 3.1868774108700002, \
        55: 3.1943478492900002, \
        56: 3.2018182877100001, \
        57: 3.20928872613, \
        58: 3.21675916455, \
        59: 3.2242296029699999, \
        60: 3.2317000413799999, \
        61: 3.2391704797999998, \
        62: 3.2466409182199998, \
        63: 3.2541113566400002, \
        64: 3.2615817950600001, \
        65: 3.2690522334800001, \
        66: 3.27652267189, \
        67: 3.28399311031, \
        68: 3.2914635487299999, \
        69: 3.2989339871499999, \
        70: 3.3064044255699998, \
        71: 3.3138748639900002, \
        72: 3.3213453024000001, \
        73: 3.3288157408200001, \
        74: 3.33628617924, \
        75: 3.34375661766, \
        76: 3.3512270560799999, \
        77: 3.3586974944999999, \
        78: 3.3661679329199998, \
        79: 3.3736383713300002, \
        80: 3.3811088097500002, \
        81: 3.3885792481700001, \
        82: 3.3960496865900001, \
        83: 3.40352012501, \
        84: 3.41099056343, \
        85: 3.4184610018399999, \
        86: 3.4259314402599998, \
        87: 3.4334018786799998, \
        88: 3.4408723171000002, \
        89: 3.4483427555200001, \
        90: 3.4558131939400001, \
        91: 3.46328363235, \
        92: 3.47075407077, \
        93: 3.4782245091899999, \
        94: 3.4856949476099999, \
        95: 3.4931653860299998, \
        96: 3.5006358244500002, \
        97: 3.5081062628600002, \
        98: 3.5155767012800001, \
        99: 3.5230471397000001, \
        100: 3.53051757812};
    if answer[0].upper() == "R":
        table = reproduce_lookup_table();
        secure_lut( table );
#END OF TRY LOAD lut.py____________________________________________________________________________
  

def extrapolate_y_shift(k): #only makes sense for k > 100
    """extrapolate_y_shift is a helper function: 
    extrapolation for k > 100 according to least squares on  values of 50,75,100,200,500
    """

    #precalculated data
    data = array([[75.,3.42],[100.,3.531],[150.,3.75],[200.,3.86],[500.,4.28]]);
    
    #data is supposed to be logarithmic in k
    for i  in xrange(len(data)):
        data[i][1] = exp(data[i][1]);
    #end for
    
    #least squares for finding parameters of linear function
    k_mean = 1./len(data) * sum(data,0)[0];
    y_mean = 1./len(data) * sum(data,0)[1];
    
    a = dot((data.T[0] - k_mean),(data.T[1] - y_mean))/norm(data.T[0] - k_mean)**2.;
    t = y_mean - a * k_mean;
    
    #print a
    #print t
    
    return log(a*k + t);
#END DEF extrapolate_y_shift_______________________________________________________________________


def adapt_error_function(k, y_shift):
    """adapt_error_function is a helper function for finding steepness parameter of Inverse Error Function"""
    
    #analytical calculation of steepness (linear transformation of ErfInv
    if not (k == 2):
        steepness = -modulated_error_function((k-1.)/k, 0, -1., 1.) / (0. - y_shift);
    else:
        #der Grund liegt in Normierung der Errorfunction: Die allgemeine Errorfunction, 
        #so wie sie mit der Aproximation im Code berechnet wird, ist anders normiert als die 
        #Gaussche Fehlerfunktion. Der Umrechnungsfaktor ist (1/sqrt(2*pi))/(2/sqrt(pi)) was 
        #ungefaehr  0.353553 ist. 
        steepness = -0.347 #better practical results
        #steepness = -0.353553 #theoretical value
    return steepness;
#END DEF adapt_error_function______________________________________________________________________

def modulated_error_function(x_values, y_center=0., steepness=1., x_width=2.):
    """Helper function. Generates a modulated inverse errorfunction with more parameters.
    returns an array. 
    
    Keyword arguments:
    x_values --	list of the x coordinates of the inverse error function
    y_center -- float. vertical center of the Errorfunction, translation parallel to y axis (Standard = 0)
    steepness -- float. steepness of the inverse Errorfunction, relative scale in y direction (Standard = 1)
    x_width -- float. absolute width of the Errorfunction, scale in x direction (Standard = 2)
    """
    
    x = array(x_values);
    y = (-1./steepness) * erfcinv( (2./x_width) * x) + y_center;   
    Result = array(y).T;
    return Result
#END DEF modulated_error_function__________________________________________________________________


def setup_simplex( n, distance ):	
    """Helper function. Creates a numpy array with coordinates of a n-simplex where any edge is >>distance<< long. 
    The return value is a 2d-numpy-array.

    Keyword arguments:
    n -- int. number of densities
    distance -- float. distance apart for densities
    """
    
    if n <= 1:
        return array([[0]])
    
    result = zeros((n, n-1))
    sub_simplex = setup_simplex(n-1, distance)
    base_foot = sub_simplex.sum(0) / (n - 1)
    r = linalg.norm(base_foot)
    y = sqrt(distance**2 - r**2)
    result[0:n-1,0:n-2] = sub_simplex
    result[n-1,0:n-2] = base_foot
    result[n-1,n-2] = y
    return result
#END DEF setup_simplex_____________________________________________________________________________


def sort_rows( data, row ):
    """Helper function. Sorts a numpy array by a given column. The return value is asorted data object."""
    data2 = array( data )
    data2 = data2[data2[:,row].argsort()]
    return data2
#END DEF sort_rows_________________________________________________________________________________


def white_gaussian( count, dims ):
    """Helper function. Creates a white gaussian. The return value is a 2d-numpy-array.
    
    Keyword arguments:		
    count -- int. how many samples.
    dims -- int. how many dimensions.    
    """
    result = []
    for i in xrange( dims ):
        result += [normal(0.0, 1.0, count)]
    result = array(result).T
    return result;
#END DEF white_gaussian____________________________________________________________________________


def gauss_likelihood( vector, center ): # inverse_epsilon, sqrt_determinant=None ): #add this if you want freely specify shape
    """Helper function. Computes the likelihood of the density specified through >>inverse_epsilon<< and
    >>sqrt_determinant<< and >>center<< at position >>vector<<. This function
    returns a lieklihood for every vector.
    
    supply >>inverse_epsilon<< and >>sqrt_determinant<< from a central place. This speeds
    up the performance:
    inverse_epsilon  = linalg.inv( Epsilon )
    sqrt_determinant = sqrt( abs( linalg.det( Epsilon ) ) )
    
    Keyword arguments:
    vector -- 2d-numpy-array which can contain a series of vectors (in rows)
    center -- 2d-numpy-array. vector containing the center position of the gaussian
    inverse_epsilon -- 2d--numpy-array. matrix with the inverse of the cov matrix
    sqrt_determinant -- float. square root of absolute determinant of the cov matrix
    """
    
    #add the following lines for more robustness
    #if type(center ) != ndarray: #this is true for 1D
    # center = array([[center]]);        

    import time;
    j = vector - center;    
    d = len( center );
    #this is the fully correct version that allows for arbitrary cov matrix
    #b = transpose(exp( -0.5 * sum(j.T * dot(inverse_epsilon, j.T), 0) ) / ((2 * pi)**(d / 2.0) * sqrt_determinant));
    b = exp( -0.5 * sum(j.T**2, 0) ) / ((2 * pi)**(d / 2.0) );
    return b;
#END DEF gauss_likelihood________________________________________________________________________


def position_whites( count, positions ):
    """Helper function. Creates a dataset with labels, positions and probabilities from white gaussians
    which are positioned at >>positions<<. The shape of >>positions<< defines
    how many dimensions are output. >>count<< defines how many samples are generated
    per density. The return values are vecs, labels, probabilities
    
    Keyword arguments:
    count -- int. number of samples per density.
    positions -- 2d-numpy-array. height of it defines number of densities and width of it defines dimensionality.

    """
    k = positions.shape[0]
    inverse_epsilon = eye(k - 1);
    determinant = 1.0
    dims = positions.shape[1]
    vectors  = white_gaussian(k*count, dims)
    #vectors = tile( white_one, (k,1))	
    index = array(range(k*count)) / count #index contains 0,0,0,0, ... 1,1,1,1,1,1,....,2,2,2,2,2,.....,k-1,k-1,k-1,k-1,.... (total length is k*count) #serves as labels
    offsets = positions[index]            #tiles positions in such a way, that the same density is positioned at each k-simplex corner
    result_vectors = vectors + offsets    #actually move densities
    probabilities = [];
    for i in xrange( k ):
        ps = gauss_likelihood(result_vectors, array([positions[i]]) ); #compute probabilities
        probabilities += ps,
        pass;
        
    p = array(probabilities).T
    sums = p.sum(1);
    p /= array([sums]).T;
    return result_vectors, index, p
#END DEF position_whites___________________________________________________________________________


#FINAL IDEA USES LOOK UP TABLES AND LOGARITHMIC EXTRAPOLATION. NO SEARCH INVOLVED
def edge_length_of_k_simplex_lut( k, error):   
    """Compute edge length for a synthetic dataset with k white gaussian densities positioned
    at the vertices of a k-simplex. The dimensionality of the data is k-1.
    target_error specifies the the datasets' Bayes error rate.
    This version uses a look up table and extrapolation until k=500.
    
    Keyword arguments:
    k -- integer. how many gaussians?
    target_error -- float. value between 0.0 and 1.0
    """
    
    k = max(k,2);
    if (k > 100): #calculate y_shift by extrapolation
        y_shift = extrapolate_y_shift(k);
    else: #lookup table for k <= 100
        y_shift = Table_y_shift[k];
    #end if
    steepness = adapt_error_function(k,y_shift);    
    dist = modulated_error_function(error, y_shift, steepness, 1.)    
    return dist;
#END DEF edge_length_of_k_simplex_lut______________________________________________________________


def edge_length_of_k_simplex_amoeba( k, error, samples_for_optimization=1e6 ):
    """Compute edge length for a synthetic dataset with k white gaussian densities positioned
    at the vertices of a k-simplex. The dimensionality of the data is k-1.
    target_error specifies the the datasets' Bayes error rate.    
    This version uses a look up table and extrapolation until k=500.
    This method finds the y_shift using fast Bayes error computation method
    and amoeba optimization technique.
    
    Keyword arguments:
    k -- integer. how many gaussians?
    target_error -- float. value between 0.0 and 1.0
    samples_for_optimization -- integer. number of samples to be used by FastWGKSErrorModel
    """
    error_model = FastWGKSErrorModel( k, 0.5, samples_for_optimization ) #samples_per_density*k );  #optimize for 50% error
    y = amoeba( error_model.cost_function, 1 )[0]
    y_shift = error_model.convert_2_true_edge_length( y )
    steepness = adapt_error_function( k, y_shift );    
    d = modulated_error_function( error, y_shift, steepness, 1.)
    return d;
#END DEF edge_length_of_k_simplex_amoeba___________________________________________________________


def edge_length_of_k_simplex_numerical_descent( k, error, samples_for_optimization=1e6 ):
    """Compute edge length for a synthetic dataset with k white gaussian densities positioned
    at the vertices of a k-simplex. The dimensionality of the data is k-1.
    target_error specifies the the datasets' Bayes error rate.    
    This version uses a look up table and extrapolation until k=500.
    This method finds the y_shift using fast Bayes error computation method
    and gradient descent optimization technique based on numerical gradients.
    
    Keyword arguments:
    k -- integer. how many gaussians?
    target_error -- float. value between 0.0 and 1.0
    samples_for_optimization -- integer. number of samples to be used by FastWGKSErrorModel
    """
    error_model = FastWGKSErrorModel( k, 0.5, samples_for_optimization ) #samples_per_density*k );  #optimize for 50% error
    y = descent( error_model.cost_function, array([1.0]) )[0]
    y_shift = error_model.convert_2_true_edge_length( y )
    steepness = adapt_error_function( k, y_shift );    
    d = modulated_error_function( error, y_shift, steepness, 1.)
    return d;
#END DEF edge_length_of_k_simplex_numerical_descent________________________________________________


def k_white_gaussians(d, k, samples_per_density):   
    """Create a synthetic dataset with k white gaussians positioned at 
    corners of a k-simplex whose edges are d apart.
    
    Keyword arguments:
    d -- float. length of edge
    k -- int. number of simplex corners
    samples_per_density -- int. number of samples per white gaussian
    """
    
    n = max(k, 2)
    positions = setup_simplex(n, d)
    return position_whites(samples_per_density, positions) #will return tripe: (v:vectors,l:labels,p:posteriors)
#END DEF k_white_gaussians_________________________________________________________________________


def white_gaussians_on_k_simplex(k, target_error, samples_per_density, optimizer="lut", samples_for_optimization=1e6 ):
    """Create a synthetic dataset with k white gaussian densities positioned
    at the vertices of a k-simplex. The dimensionality of the data is k-1.
    target_error specifies the the datasets' Bayes error rate and 
    samples_per_density define how many values are sample from any gaussian.
    It is a conveniency function wrapping edge_length_of_k_simplex and k_white_gaussians.
    
    Keyword arguments:
    k -- integer. how many gaussians?
    target_error -- float. value between 0.0 and 1.0
    samples_per_density -- integer. how many samples per gaussian?
    sigma -- float. spreading constant for cross-entropy (default=10.0)
    elite_size -- int. how many elites are taken from the population for seeding next population (default=20)
    values -- int. size of cross-entropy population (default=100)
    """
    if optimizer.lower() == "lut":
        d = edge_length_of_k_simplex_lut(k, target_error );
    if optimizer.lower() == "amoeba":
        d = edge_length_of_k_simplex_amoeba( k, target_error, samples_for_optimization );
    if optimizer.lower() == "descent":
        d = edge_length_of_k_simplex_numerical_descent( k, target_error, samples_for_optimization );

    return k_white_gaussians( d, k, samples_per_density );
#END DEF white_gaussians_on_k_simplex______________________________________________________________











#==================================================================================================
#TEST CASES
#==================================================================================================
"""The following section defines test cases"""
if __name__ == "__main__":
    from time import *;

    print "TEST d ESTIMATION:"
    k = 3
    target_error = 0.3
    samples_for_optimization = 1e5
    print "LUT: ", edge_length_of_k_simplex_lut(k, target_error );
    print "AMOEBA ON FASTWGKS: ", edge_length_of_k_simplex_amoeba( k, target_error, samples_for_optimization );
    #print "DESCENT ON FASTWGKS: ", edge_length_of_k_simplex_numerical_descent( k, target_error, samples_for_optimization );
    exit();

    
    print "perforiming selftest of module"
    v,l,p = white_gaussians_on_k_simplex( 10, 0.3, 100, "descent" )

    
    #test case 1
    test_res = array([[0.0,0.0,0.0 ],[3.0,0.0,0.0],[1.5,2.59807621,0.0],[1.5,0.8660254,2.44948974]]);
    x = setup_simplex(4, 3.0);
    if sum(x - test_res) > 0.001:
        raise Exception( "problem with setup_simplex" )
    else: print "setup_simplex passed!"

    
    #test case 2
    sorted_res = array([[0.0,0.0,0.0],[1.5,2.59807621,0.0],[1.5,0.8660254,2.44948974],[3.0,0.0,0.0]])
    x = sort_rows(test_res, 0)
    if sum(x - sorted_res) > 0.001:
        print "if this test failes, then try one more time"
        raise Exception( "problem with test_res" )
    else: print "sort_rows passed!"

    #test case 3
    x = white_gaussian(10000, 20);
    if x.shape != (10000, 20):
        raise Exception( "problem with white_gaussian (1)" )
    if sum(mean(x, 0) - zeros(20) ) > 0.3:
        print "if this test failes, then try one more time"
        print sum( mean(x, 0) - zeros(20) )
        raise Exception( "problem with white_gaussian (2)" )	
    if sum(cov(x.T) - eye(20) ) > 0.3:
        print "if this test failes, then try one more time"
        print sum( cov(x.T) - eye(20) )
        raise Exception( "problem with white_gaussian (3)" )	
    print "white_gaussian passed"
    
    #test case 4
    x = rand( 10, 2);
    y = gauss_likelihood( x, array([[0,0]]) )
    if len( y ) != 10:
        raise Exception( "problem with gauss_likelihood!" )
    print "gauss_likelihood passed!"

    #test case 5
    x, l, p = position_whites( 10, array([[0,0],[10,0],[0,10]]) )
    if x.shape != (30,2):
        raise Exception( "problem with position_whites (1)" )    
    if len( l ) != 30:
        raise Exception( "problem with position_whites (2)" )
    print "position_whites passed!"
            
    v, l, p = white_gaussians_on_k_simplex( 20, 0.1, 100 )
    if v.shape != (2000,19):
        print v.shape
        raise Exception( "problem with white_gaussians_on_k_simplex (1)" )    
    if len( l ) != 2000:
        print len( l ) 
        raise Exception( "problem with white_gaussians_on_k_simplex (2)" )
    print "white_gaussians_on_k_simplex passed!"


#==================================================================================================
#END OF FILE
#==================================================================================================

