#==================================================================================================
#White Markov Chain Generator WMCG
#Authors: Aleksander Lodwich, Janick Frasch and Thomas M. Breuel
#Remark: from the PaREn Project. 
#Refer to "Practical Bayes-True Data Generators For Evaluation of 
#          Machine Learning, Pattern Recognition and Data Mining Methods"
#
#This code is under MIT licence.
#Copyright (c) 2009, DFKI
#
#Permission is hereby granted, free of charge, to any person obtaining a copy of 
#this software and associated documentation files (the "Software"), to deal in 
#the Software without restriction, including without limitation the rights to use, 
#copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
#Software, and to permit persons to whom the Software is furnished to do so, 
#subject to the following conditions:
#
#The above copyright notice and this permission notice shall be included in all 
#copies or substantial portions of the Software.
#
#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS 
#FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
#COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
#IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#==================================================================================================

#==================================================================================================
#INCLUDES
#==================================================================================================
from time import *
from pylab import *
from optimizers import *
from numpy import *
from scipy import *
import sys


ENABLE_VISUALIZATION = 0  #0 is default

#==================================================================================================
#FUNCTIONAL DEPENDENCIES:
#==================================================================================================
#
#                  gcg()                            wmcg()
#                   |                                 |
#                  / \                               / \
#                 /   \                             /   \
#                /     -- CLASS SkelettonSolution --     \
#                |                            |           |
#                |bayes_error_method_a/b()----+           |
#                |                            |           |
#                |        wmcg_posteriors()---+           |
#                |                            |           |
#                |    flesh_on_skeleton()-----+           |
#                |            |                           |
#                |         density_tube()                 |
#                |                                        |
#         gaussian_skeletton()                   grow_wmcg_skeletton()
#                |                                        |
#                |\                                      /|
#                | -----------gaussian()----------------- |
#                |                                        |
#          grow_skeletton()                              /|
#                |                                      / |
#                |\                                    / /
#                | ---------- sort_rows()-------------- /
#                 \                                    /
#                  ---------array_without_rows()-------


#==================================================================================================
#FUNCTIONS
#==================================================================================================
def sort_rows( data, row ):
    """Helper function. Sorts a numpy array by a given column. The return value is asorted data object."""
    data2 = array( data )
    data2 = data2[data2[:,row].argsort()]
    return data2
#END DEF sort_rows_________________________________________________________________________________


def array_without_rows( data, rows ):
    """Helper function. data is at least 2d numpy array and rows is a list of indices for rows
    to be extracted. At the same time the remaining rows are also returned in a separate object.
    """
    remaining = arange( 1, len(data)+1 )
    remaining[rows] = 0
    remaining = unique( remaining )[1:] - 1
    return data[rows], data[remaining]
#END DEF array_without_rows________________________________________________________________________


def get_indices( bool_vector ):
    return unique( concatenate( (array([0]), bool_vector * arange( 1, len(bool_vector)+1 ) ) ) )[1:]-1
#END DEF get_indices_______________________________________________________________________________    
        

def gaussian( count, sigma ):
    """Basic function. Returns samples from a multivariate gaussian density.

    Keyword Arguments:
    count:  int>0: how many samples do you need?
    sigma:  2d square array: what is the covariance matrix?
    """
    dims = sigma.shape[0]

    result = []
    for i in xrange( dims ):
        result += [random.normal(0.0, 1.0, count)]
    result = array(result).T

    amp, vecs = linalg.eig( sigma )
    transform = (vecs.T*amp)
    result = dot( transform, result.T ).T
    return result;
#END DEF gaussian__________________________________________________________________________________


def grow_skeletton( vectors, no_of_chains, length, wished_angle=0, speed=1e-3 ):
    """grow_skeletton grows skelettons from a pool of predefined allowed vertices supplied by vectors.
    The vectors can be provided by any distribution of one own's like. The only
    restriction is that the number of provided vectors must be large than no_of_chains*length.
    This function smooths the created curves according to wished_angle and speed.
    The smaller the angle (expressed in degrees) and the higher the speed,
    the smoother the curves. The purpose of this growing function is to control the
    areas of high path crossing.

    Keyword Arguments:
    vectors:        2d array of size=(random,dims). Prototype vertices.
    no_of_chains:   int>1: number of chains to be created
    length:         int>1: number of vertices to be created for every chain
    wished_angle:   float [0,180]: prefered level of curvature for every vertex
    speed:          float>0.0: regularization factor for trading off between closeness and angle of adjacent path sections.
    """
    angle = cos( wished_angle * pi/180 )
    v = copy( vectors )
    if len( v ) < no_of_chains * length:
        msg = "ERROR ON " + ctime() + " grow_skeletton had " + str( len( v ) ) + " seed vectors, but needed at least " + str( no_of_chains * length ) + "\n"
        log_file = open( "skeletton_errors.log", "a") 
        log_file.write( msg )
        log_file.close()
        raise Exception( msg )

    skeletton = zeros( (no_of_chains, vectors.shape[1], length ) )
    indices = random.permutation( arange( len( v ) ) )[:no_of_chains] 
    skeletton[:,:,0], v = array_without_rows( v, indices )

    #this for loops does a nearest neighbor search to the points in skeletton[:,:,0]
    for n in arange( no_of_chains ):
        new_dirs = v - skeletton[n,:,0]
        new_norms = sqrt(sum((new_dirs**2),1))[:,newaxis]
        indices = arange( len( v ) )[:,newaxis]
        choice_matrix = sort_rows( concatenate( (new_norms, indices, v ), 1 ), 0 ) #sort by norms
        skeletton[n,:,1], v = array_without_rows( choice_matrix[:,2:], array([0]) )#take closest so far        

    for d in arange( 2, length ):
        #compute the direction vectors for every snake
        last_dirs = skeletton[:,:,d-1] - skeletton[:,:,d-2]
        last_norms = sqrt(sum((last_dirs**2),1))
        last_dirs/=last_norms[:,newaxis]
        for n in arange( no_of_chains ): #for every snake do the following:
            #1. estimate directions to all yet possible points in v
            new_dirs = v - skeletton[n,:,d-1]
            new_norms = sqrt(sum((new_dirs**2),1))
            new_dirs /= new_norms[:,newaxis]
            #2. compute a cost function based on "speed" and the angle
            costs = ( sum(last_dirs[n,:]*new_dirs,1)[:,newaxis] - angle )**2 + ( new_norms / speed )[:,newaxis]
            indices = arange( len( v ) )[:,newaxis]
            #3. pick vector that has the lowest cost
            choice_matrix = sort_rows( concatenate( (costs, indices, v ), 1 ), 0 )
            skeletton[n,:,d], v = array_without_rows( choice_matrix[:,2:], array([0]) ) #this will reduce v in size
    return skeletton
#END DEF grow_skeletton____________________________________________________________________________


def grow_wmcg_skeletton( chain_seed_vectors, length, wished_angle=0, speed=0.1, test_samples=100, fix_speed=0, fix_angle=0, resample=1 ):
    """grow_wmcg_skeletton. This function will create a 3D array of values where 1st dim is the number of paths, where 2nd dim is
    the number of features and 3rd dim is the number of path vertices. The bending and distribution of the chains can be controlled
    by the following parameters:

    Keyword Arguments:
    chain_seed_vectors:     2d array of size=(no_of_chains,no_of_dims). Contains starting points for every chain.
    length:                 int>0. How many vertices shall be generated for every chain?
    wished_angle:           float [0.0,180.0]. Depending on fix_angle is interpreted in the following two ways:
                                fix_angle=1: find random directional vectors that are most closely different by this angle from the current direction
                                fix_angle=0: find random directional vectors that are at most different by this angle from the current direction
    speed:                  float>0.0: scales the length of each chain component by this factor.
    test_samples:           int>0: defines how many directions are sampled. The more the finer the model
    fix_speed:              0/1: if 1 then all chains' length=speed
    fix_angle:              0/1: if 1 then angle between current direction and new direction is defined by wished_angle
    resample:               0/1: if 1 then test samples will be resampled before every new vertex is computed
    """

    angle              = cos( wished_angle * pi/180 ) #transfer degree values into radial value
    dims               = chain_seed_vectors.shape[1]  #how many feature dimensions do we have?
    vector_source      = gaussian( test_samples, eye( dims )) #generate a prototype of all directions and speeds
    vector_source_norm = vector_source * sqrt( sum( vector_source**2, 1 ) )[:,newaxis] #only directions and speed =1
    vector_source_adj  = vector_source_norm * speed; #uniformely distributed sphere of directions and speed=speed
    indices            = arange( test_samples )
    no_of_paths        = chain_seed_vectors.shape[0] #how many paths are requested?

    #create the first two vertices of any path:
    skeletton          = zeros( (no_of_paths, dims, length ) )
    skeletton[:,:,0]   = chain_seed_vectors;
    skeletton[:,:,1]   = chain_seed_vectors + vector_source_adj[random.permutation(indices)[:no_of_paths]];
 
    #create the rest of the vertices of any path:
    for d in arange( 2, length ):
        vector_source      = gaussian( test_samples, eye( dims ) ) #white density with normally distributed speeds and uniformely distributed directions
        vector_source_norm = vector_source / sqrt( sum( vector_source**2, 1 ) )[:,newaxis] #speeds are constant
        if fix_speed:
            vector_source_adj  = vector_source_norm * speed; #uniformely distributed sphere of directions with speed=speed
        else:
            vector_source_adj  = vector_source * speed; #uniformely distributed sphere of directions with speed=N()*speed

        #compute the direction vectors for every snake
        last_dirs = ( skeletton[:,:,d-1] - skeletton[:,:,d-2] )
        last_dirs /= sqrt( sum( last_dirs**2, 1 ) )[:,newaxis] #we get unit vectors, this is important for the dot product later
        for n in arange( no_of_paths ): #for every snake do the following:
            #1. estimate directions to all yet possible points in v
            new_dirs = vector_source_norm #that's cheap
            #2. compute a cost function based on "speed" and the angle
            if not fix_angle:
                #accept angle in range [angle,1]
                costs = (maximum( 0, angle - sum(last_dirs[n,:]*new_dirs,1) ) + ( rand( test_samples ) * 1e-3 ))[:,newaxis] #make sure there's enough randomness for sort
            else:
                #accept angles ~ [angle-noise,angle+noise]
                costs = ( sum(last_dirs[n,:]*new_dirs,1)[:,newaxis] - angle )**2

            #3. pick vector that has the lowest cost
            choice_matrix = sort_rows( concatenate( (costs, indices[:,newaxis], vector_source_adj+skeletton[n,:,d-1] ), 1 ), 0 )

            skeletton[n,:,d], garbage = array_without_rows( choice_matrix[:,2:], array([0]) ) #only pick the best vector array([0])=0
    return skeletton
#END DEF grow_wmcg_skeletton_______________________________________________________________________


def gaussian_skeletton( count, sigma, no_of_chains, length, wished_angle=0, speed=1e-3 ):
    """gaussian_skeletton. This function generates a skeletton from a prototype normal
    distribution. The number of crossings of the chains increases in the center and
    decreases outwards.

    Keyword Arguments:
    count:          int>0: how many prototypes to use?
    sigma:          2d square array: what is the covariance matrix of the prototype distribution?
    no_of_chains:   int>1: how many chains shall be generated?
    length:         int>1: how many vertices shall exist in every chain?
    wished_angle:   float[0,180]: what is the target angle between two adjacent path sections?
    speed:          float>0:  what is the regularization factor for the tradeoff between angle and vertex distance?
    """
    g = gaussian( count, sigma )
    return grow_skeletton( g, no_of_chains, length+1, wished_angle, speed )
#END DEF gaussian_skeletton________________________________________________________________________


def density_tube( length, dims, count, sigma ):
    """Helper Function. density_tube generates a "tube" of sphericaly distributed samples
    around dim 0 for the length of length. The density along dim0 is controlled via
    longitudianl_density and sigma controlls the "thickness" of the tube. The tube
    always starts at vec=0. The tube has to be transformed to the proper place later.

    Keyword Arguements:
    length:     float>0.0: how lang shall the tube be?
    dims:       int>0: how many dimensions does the problem have?
    count:      int>0: how many samples shall be generated?
    sigma:      float>0: how "wide" is the tube? Real width is infinite of course.
    """
    result = zeros( ( count, dims ) )
    result[:,1:] = random.normal( scale=sigma, size = ( count, dims-1 ) ) #first dim must remain zero
    if count > 0:
        result[:,0] = rand( count ) * length #stretch samples along length
    return result
#END DEF density_tube______________________________________________________________________________


def flesh_on_skeletton( skeletton, total_count, sigma ):
    """flesh_on_skeletton is one of the core functionality. It basically puts samples
    around the skeletton paths and delivers a new dataset with ground truth labels.

    Keyword Arguments:
    skeletton:              3d array of size=(no_of_chains,no_of_dims,chain_length): as can be grown using grow_skeletton()
    total_count:            int>>1. How many samples shall be generated?
    sigma:                  float>0.0. How "wide" are the chains?
    """
    sigma = max( 1e-10, sigma )
    dims = skeletton.shape[1]
    sections = skeletton.shape[2]-1
    snakes = skeletton.shape[0]

    #compute total length and create a sampling plan for the
    total_length = 0
    lengths = zeros( snakes * sections ) #segment probabilities
    seg_p   = zeros( snakes * sections ) #segment probabilities
    for n in arange( snakes ):
        for section in arange( sections ):
            path_vec = skeletton[n,:,section+1]-skeletton[n,:,section][newaxis,:]
            length = linalg.norm( path_vec )
            i = n * sections + section
            lengths[i] = length
    seg_p = (lengths / sum(lengths))
    samples = ( seg_p * total_count ).round()
    samples[0] += total_count - sum(samples) #stuff missing samples

    #in this section we'll generate the samples around the skeletton
    vectors = {}
    for n in arange( snakes ):
        for section in arange( sections ):
            i = n * sections + section
            path_vec = skeletton[n,:,section+1]-skeletton[n,:,section][newaxis,:]
            tube = density_tube( 1.0, dims, samples[i], sigma )
            V = linalg.svd(path_vec)[2]            
            V[0] = path_vec #assure that the path vector is oriented correctly
            transformation = V.T
            aligned_tube = dot( transformation, tube.T ).T + skeletton[n,:,section][newaxis,:]
            vectors[n] = concatenate( ( vectors.get( n, zeros((0,dims))), aligned_tube ) )

    #in this ection we'll compact the dictionary into a single dataset  
    #TODO: we're growing result and labels which can be somewhat inefficient!  
    result = zeros( (0, dims) )
    labels = zeros( 0 )
    label_index = 0
    for v in vectors.values():
        result = concatenate( ( result, v ) )
        labels = concatenate( ( labels, ones( len( v ) )*label_index ) )
        label_index += 1  

    return result, labels
#END DEF flesh_on_skeletton________________________________________________________________________


def wmcg_posteriors( samples, skeletton, sigma ):
    """wmcg_posteriors computes the posterior probabilities for a wmcg model based on the skeletton
    and the sigma. Posteriors are computed for every vector in samples.

    Keyword Arguments:
    samples:    2d array. vectors are rows
    skeletton:  3d array. dim0: chains dim1:features dim2:vertices of chains
    sigma:      float>0. width of chains
    """
    paths  = skeletton.shape[0]
    joints = skeletton.shape[2]-1
    dims   = skeletton.shape[1]
    sample_count = samples.shape[0]

    posteriors = zeros( (sample_count,paths) )
    #for vec_index in arange( samples ):
    for path_index in arange( paths ):
        for joint_index in arange( joints ):
            step_vector   = skeletton[path_index,:,joint_index+1] - skeletton[path_index,:,joint_index]
            length        = linalg.norm( step_vector )
            direction     = ( step_vector / length )[newaxis,:]
            local_samples = copy( samples ) - skeletton[path_index,:,joint_index][newaxis,:]
            dot_products  = sum( local_samples * direction, 1 )
            involved_vecs = float32( ( dot_products >= 0.0 ) & ( dot_products <= length ) )
            hypotenuse    = sum( local_samples**2 ,1) #entf**2 is already squared
            distances     = sqrt( hypotenuse - dot_products**2 ) #therefore you need to do it here
            likelihoods   = exp( -0.5 * distances**2 / sigma )
            posteriors[:,path_index] += likelihoods * involved_vecs   #at first these are likelihoods        
    #end for
    
    posteriors /= sum( posteriors, 1 )[:,newaxis]+1e-308 #now these are posteriors
    return posteriors;
#END DEF wmcg_posteriors___________________________________________________________________________


def bayes_error_method_a( l1, l2 ):
    """computes the Bayes error by comparing l1 (ground truth) and l2 (estimated labels)"""
    return 1.0 - mean( l1 == l2 )
#END DEF bayes_error_method_a______________________________________________________________________


def bayes_error_method_b( probabilities ):
    """computes the Bayes error by probability mass integration."""
    return 1.0 - mean( probabilities.max(1) ) #compute Bayes error according to method (B)
#END DEF bayes_error_method_b______________________________________________________________________



if ENABLE_VISUALIZATION:
    glob_fig = None
    def drawt( s, v, l, y_lim=[-10,10], x_lim=[-10,10], wait=3 ):
        """drawt is a DrawTranslator() function replacement"""
        global glob_fig
        ion()
        if glob_fig is None:
            glob_fig = figure(figsize=(15,15))
        glob_fig.clear()
        glob_fig.show()
        d = len(s)
        hold( 'on' )
        for k in range( d ):
            selected = get_indices( int32( l == k ) )
            plot( v[selected,0], v[selected,1], '.' )        
        for k in range( d ):
            plot( s[k,0,:], s[k,1,:] )
            plot( s[k,0,:], s[k,1,:],'o' )
        ylim(y_lim);xlim(x_lim)
        hold( 'off' )
        sleep(wait)
#END DEF drawt_____________________________________________________________________________________


class SkelettonSolution:
    """SkelettonSolution is a cookie class for keeping all the extra constants
    needed to compute the Bayes error in the cost_function.
    """

    def __init__( self, skeletton, target_error, total_samples=1000 ):
        self.my_skeletton = skeletton
        self.my_target_error = target_error
        self.my_total_samples = total_samples
        self.my_target_samples_count = 10000 #tmp
        result = amoeba( self.cost_function, 1, initial_values=array([[0.9],[1.1]]), abort_under_cost=1e-4 )
        if abs( self.my_bayes_error - target_error ) > 0.02:
            msg = "ERROR ON " + ctime() + " target error was missed. target is " + str(target_error) + " and achieved was " + str( self.my_bayes_error ) + "\n"
            log_file = open( "skeletton_errors.log", "a" )
            log_file.write( msg )
            log_file.close()
            raise Exception( msg )
        sys.stdout.flush()
    #END DEF __INIT__
   

    def cost_function( self, sigma_in_list ):
        """the form of the function is required by the optimizers and cannot be changed"""
        self.my_sigma = sigma_in_list[0]
        if self.my_sigma <= 0.0:
            return 1.0;
        #else:
        self.my_samples, self.my_labels = flesh_on_skeletton( self.my_skeletton, self.my_total_samples, self.my_sigma )
        self.my_posteriors = wmcg_posteriors( self.my_samples, self.my_skeletton, self.my_sigma )
        self.my_bayes_error = bayes_error_method_a( self.my_labels, argmax( self.my_posteriors,1 ) )
        #DEBUG SECTION
        if ENABLE_VISUALIZATION:
            drawt( self.my_skeletton, self.my_samples, self.my_labels )
            print "[sigma,total_samples,samples,error]=", self.my_sigma, self.my_total_samples, len(self.my_samples), self.my_bayes_error
            print "cost of error:", ( self.my_target_error - self.my_bayes_error )**2
        #CREATE REFERENCE VARIABLES FOR CONVENIENCE IN gnerators.py AND THEY DO NOT COST EXTRA MEMORY
        self.s = self.my_skeletton
        self.v = self.my_samples
        self.l = self.my_labels
        self.p = self.my_posteriors
        return 100*( self.my_target_error - self.my_bayes_error )**2 
    #END DEF cost_function


    def __str__( self ):
        return "error:" + str(self.my_bayes_error) + " at sigma:" + str(self.my_sigma)
    #END DEF __STR__
#END DEF CLASS SkelettonSolution___________________________________________________________________




def gcg( target_error, seeds=150, seed_sigma=eye(2)*5, no_of_chains=10, chain_lengths=15, target_angle=0, speed=1e-0, total_samples=1000 ):
    """gcg generates a gaussian chain generator dataset.

    Keyword Arguments:
    target_error:       float[0,1]. what shall the Bayes error for the dataset be?
    seeds:              int>>1. how many seed starting points shall be generated on a normal distribution?
    seed_sigma:         2d square array. covariance matrix for the normal seed distribution
    no_of_chains:       int>1. how many chains shall be generated?
    chain_lengths:      int>1. how many vertices shall every chain have?
    target_angle:       float[0,180]. what is the angle between adjacent path sections?
    speed:              float>0. regularization parameter for making curves smoother
    total_samples:      int>>1. What is the number of vectors you want to generate?
    """
    s = gaussian_skeletton( seeds, seed_sigma, no_of_chains, chain_lengths, target_angle, speed )
    result = SkelettonSolution( s, target_error, total_samples )
    return s, result.my_samples, result.my_labels, result.my_posteriors
#END DEF gcg_______________________________________________________________________________________


def scg( target_error, radius=10.0, seeds=150, seed_sigma=eye(3)*5, no_of_chains=10, chain_lengths=10, target_angle=0, speed=1e-0, total_samples=1000 ):
    """gcg generates a spheric chain generator dataset. Around an spheric-like structure 
    you have highest sample density and the core is rather empty.

    Keyword Arguments:
    target_error:       float[0,1]. what shall the Bayes error for the dataset be?
    seeds:              int>>1. how many seed starting points shall be generated on a normal distribution?
    seed_sigma:         2d square array. covariance matrix for the normal seed distribution
    no_of_chains:       int>1. how many chains shall be generated?
    chain_lengths:      int>1. how many vertices shall every chain have?
    target_angle:       float[0,180]. what is the angle between adjacent path sections?
    speed:              float>0. regularization parameter for making curves smoother
    total_samples:      int>>1. What is the number of vectors you want to generate?
    """
    s = gaussian_skeletton( seeds, seed_sigma, no_of_chains, chain_lengths, target_angle, speed )
    s *= radius/sqrt( sum( s**2, 1 ) )[:,newaxis,:] #normalize
    result = SkelettonSolution( s, target_error, total_samples )
    return s, result.my_samples, result.my_labels, result.my_posteriors
#END DEF gcg_______________________________________________________________________________________


def wmcg( target_error, seed_sigma=eye(2)*2, no_of_chains=10, chain_lengths=10, target_angle=180.0, 
            speed=1e-0, test_samples=1000, total_samples=1000,
            fix_speed=0, fix_angle=0, resample=1 ):
    """wmcg is a function that generates white markov chain datasets with a specifiec Bayes error rate.
    It can control the angles of the random process. Seit target_angle to 180.0 deg in order to obtain
    a Wiener process.

    Keyword Arguments:
    target_error:       float[0,1]. The target Bayes error rate of the new dataset.
    seed_sigma:         2d square array. Covariance matrix of the seeds for the chains
    no_of_chains:       int>1. How many chains shall be generated?
    chain_lengths:      int>1. How many chain-vertices shall be generated (length of chain)?
    target_angle:       float[0,180]. What is the cone of angle restriction?
    speed:              float>0.0. scaling factor for the chains
    test_samples:       int>>1. number of samples for direction sampling. low values cause crude direction changes.
    total_samples:      int>>1. What is the number of vectors you want to generate?
    fix_speed:          0/1. If yes(1), then chain sections are of equal length
    fix_angle:          0/1. If yes(1), then angle lie on the angle-cone    
    """
    g = gaussian( no_of_chains, seed_sigma )
    s = grow_wmcg_skeletton( g, chain_lengths, target_angle, speed, test_samples, fix_speed, fix_angle, resample )
    result = SkelettonSolution( s, target_error, total_samples )
    return s, result.my_samples, result.my_labels, result.my_posteriors
#END DEF wmcg______________________________________________________________________________________









#==================================================================================================
#TEST CASES
#==================================================================================================
""" Test Cases 
from PaREn import *

def debug_function_1():
    d = 10
    g = gaussian( 1000, eye(2) )
    s = grow_skeletton( g, d, 40, 0.0, 1e-3 )

    plot(  g[:,0], g[:,1], "." )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
    ylim([-5,5]);xlim([-5,5])
    show()

def debug_function_3():
    g = density_tube( 1.0, 2, 100.0, 0.1 )
    path_vec = array([[-3,1]])
    V = svd(path_vec)[2]            
    V[0] = path_vec #assure that the path vector is oriented correctly
    transformation = V.T
    g = dot( transformation, g.T ).T + array([10,0])
    plot(  g[:,0], g[:,1], "." )
    show()

def debug_function_2():
    d = 3
    g = gaussian( 1500, eye(2)*5 )
    s = grow_skeletton( g, d, 500, 20.0, 1e+1 )
    v,l = flesh_on_skeletton( s, 100.0, 0.1 )
    t = Translator( v )
    t.SetLabel( l.tolist() )
    print l
    DrawTranslator( t )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
        plot( s[k,0,:], s[k,1,:],'x' )
    #ylim([-2,2]);xlim([-2,2])
    sleep( 10 )


def debug_function_4():
    d = 10
    sigma = 0.05
    g = gaussian( 1500, eye(2)*2 )
    s = grow_skeletton( g, d, 60, 0.0, 1e+1 )
    if 0: #if mesh version
        X,Y = meshgrid( frange(-10,10,0.1), frange(-10,10,0.1) )
        x = X.flatten()
        y = Y.flatten()
        v = concatenate( (x[:,newaxis],y[:,newaxis]), 1 )
    else:
        v,l = flesh_on_skeletton( s, 100.0, sigma )    
    p = wmcg_posteriors( v, s, sigma )
    l2 = argmax( p, 1 )
    print "Bayes Error:", bayes_error_method_a( l, l2 )
    t = Translator( v )
    t.SetLabel( l2.tolist() )
    figure( figsize=(15,15) )
    DrawTranslator( t )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
        plot( s[k,0,:], s[k,1,:],'o' )
    ylim([-10,10]);xlim([-10,10])
    
    sleep( 10 )

def debug_function_5():
    s = gaussian_skeletton( 150, eye(2)*2, 10, 5, 0.0, 1e-1 )
    x = SkelettonSolution( s, 0.3 )
    print x


def debug_function_6():
    d=10
    s, v,l,p = gcg( 0.05,no_of_chains=d )
    t = Translator(v)
    t.SetLabel( l.tolist() )
    figure( figsize=(15,15))
    DrawTranslator( t )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
        plot( s[k,0,:], s[k,1,:],'o' )
    ylim([-10,10]);xlim([-10,10])
    sleep(10)


def debug_function_7():
    d = 3
    g = gaussian( 5, eye(2)*2 )
    s = grow_wmcg_skeletton( g, 15, 45.0, 1e+0 )
    v,l = flesh_on_skeletton( s, 100.0, 0.01 )
    t = Translator( v )
    t.SetLabel( l.tolist() )
    DrawTranslator( t )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
        plot( s[k,0,:], s[k,1,:],'x' )
    #ylim([-2,2]);xlim([-2,2])
    sleep( 10 )


def debug_function_8():
    d = 10
    s, v,l,p = wmcg( 0.25, seed_sigma=eye(2)*2, no_of_chains=d, chain_lengths=15, target_angle=45.0, test_samples=15, total_samples=1050 )
    t = Translator(v)
    t.SetLabel( l.tolist() )
    figure( figsize=(15,15))
    DrawTranslator( t )
    for k in range( d ):
        plot( s[k,0,:], s[k,1,:] )
        plot( s[k,0,:], s[k,1,:],'o' )
    ylim([-10,10]);xlim([-10,10])
    sleep(10)

debug_function_1()
debug_function_2()
debug_function_3()
debug_function_4()
debug_function_5()
debug_function_6()
debug_function_7()
debug_function_8()
"""
